# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.core.manifest`
===========================

This module implemenets parsing and representing the manifest files of
Craftr packages. Every Craftr package also provides a Craftr module which
can be loaded with the ``require()`` function.

Example manifest:

.. code:: json

  {
    "name": "username.packagename",
    "version": "1.0.3",
    "main": "Craftrfile",
    "author": "User Name <username@nameuser.org>",
    "url": "https://github.com/username/packagename",
    "dependencies": {
      "another_user.another_package": "1.x"
    },
    "options": {
      "directory": "path",
      "version": "string"
    }
  }
"""

from craftr.core.logging import logger
from craftr.utils import httputils
from craftr.utils import path
from craftr.utils import pyutils
from nr.types.recordclass import recordclass
from nr.types.version import Version, VersionCriteria

import abc
import fnmatch
import io
import json
import jsonschema
import nr.misc.archive
import re
import string
import urllib.request


def validate_package_name(name):
  """
  Validates the package *name*. A valid package name must consist only
  of letters, digits, hyphes, underscores and periods. The first character
  of a package must be a letter or digit.

  :raise ValueError: If *name* is not a valid package name.
  """

  if not isinstance(name, str):
    raise TypeError("package name must be str but got {!r}".format(tn(name)))
  if not re.match(r'^[A-z0-9][A-z0-9\.\-\_]*$', name):
    raise ValueError("invalid package name: {!r}".format(name))


class Namespace(object):
  """
  An empty class which is used to assign arbitrary attributes to. An instance
  of this class is created by :meth:`Manifest.get_options_namespace`.
  """

  def __str__(self):
    attrs = ', '.join('{}={!r}'.format(k, v) for k, v in vars(self).items())
    return 'Namespace({})'.format(attrs)


class InvalidManifest(Exception):
  """
  This exception can be raised by :meth:`Manifest.parse`
  """


class Manifest(recordclass):
  """
  Represents the manifest of a Craftr package. The manifest contains the basic
  information about the package name, version and its dependencies, as well as
  its author and homepage URL.

  .. attribute:: name

    The name of the Craftr package.

  .. attribute:: version

    The version of the Craftr package. This must be a semantic version
    number parsable by the :class:`Version` class.

  .. attribute:: main

    The name of the main build script that is executed for the package.
    If this member is None, Craftr will interpret it as the default value
    specified in the :class:`craftr.core.session.Session`.

  .. attribute:: project_dir

    Relative path to the project directory. Local paths in the build script
    are automatically assumed relative to this directory. Defaults to ``..``.

  .. attribute:: author

    Name of the package author.

  .. attribute:: url

    The homepage of the project of this package.

  .. attribute:: dependencies

    A dictionary that maps package names to version critiera parsable by
    the :class:`VersionCriteria` class. The criteria is very similar to
    Node.js version selectors.

  .. attribute:: options

    A dictionary of options that can be provided by the user before
    Craftr is being executed. The option name maps to a :class:`BaseOption`
    instance.
  """

  Schema = {
    "type": "object",
    "required": ["name", "version"],
    "properties": {
      "name": {"type": "string"},
      "version": {"type": "string"},
      "main": {"type": "string"},
      "project_dir": {"type": "string"},
      "author": {"type": "string"},
      "url": {"type": "string"},
      "dependencies": {
        "type": "object",
        "additionalProperties": {"type": "string"}
      },
      "options": {
        "type": "object",
        "additionalProperties": {
          "oneOf": [
            # Plain type description
            {
              "type": "string"
            },
            # Detailed option description
            {
              "type": "object",
              "properties": {
                "type": {"type": "string"}
              },
              "additionalProperties": True
            }
          ]
        }
      }
    }
  }

  Invalid = InvalidManifest

  __slots__ = tuple(Schema['properties'].keys())

  def __init__(self, name, version, main='Craftrfile', project_dir='.',
               author=None, url=None, dependencies=None, options=None):
    if version is not None:
      version = Version(version)
    self.name = name
    self.version = version
    self.main = main
    self.project_dir = project_dir
    self.author = author
    self.url = url
    self.dependencies = dependencies or {}
    self.options = options or {}

  def get_options_namespace(self, provider, errors=None):
    """
    Create a :class:`Namespace` object filled with the option values specified
    in the manifest where the option values are read from *provider*.

    :param provider: A dictionary that provides option values.
    :param errors: A list to which tuples of error information will be
      appended. The tuples are in the format (option, value, exc).
    :return Namespace
    """

    if errors is None:
      errors = []
    ns = Namespace()
    for key, option in self.options.items():
      value = provider.get(self.name + '.' + option.name, NotImplemented)
      if value is NotImplemented and option.inherit:
        value = provider.get(option.name)
      if value is NotImplemented or value is None:
        value = option.default
      else:
        try:
          value = option(value)
        except ValueError as exc:
          errors.append((option, value, exc))
          value = option.default
      setattr(ns, key, value)
    return ns

  @staticmethod
  def parse_string(string):
    """
    Shortcut for wrapping *string* in a file-like object and padding it to
    :func:`parse`.
    """

    return Manifest.parse(io.StringIO(string))

  @staticmethod
  def parse(file):
    """
    Parses a manifest file and returns a new :class:`Manifest` object.

    :param file: A filename or file-like object in JSON format.
    :raise InvalidManifest: If the file is not a valid JSON file or the
      manifest data is invalid or inconsistent.
    :return: A :class:`Manifest` object.
    """

    try:
      if isinstance(file, str):
        with open(file) as fp:
          data = json.load(fp)
      else:
        data = json.load(file)
      jsonschema.validate(data, Manifest.Schema)
    except (json.JSONDecodeError, jsonschema.ValidationError) as exc:
      raise Manifest.Invalid(exc)

    try:
      validate_package_name(data['name'])
    except ValueError:
      raise Manifest.Invalid("invalid package name: {!r}".format(data['name']))

    data.setdefault('dependencies', {})
    for key, value in data['dependencies'].items():
      try:
        data['dependencies'][key] = VersionCriteria(value)
      except ValueError as exc:
        msg = 'invalid version critiera for {0!r}: {1!r}'
        raise Manifest.Invalid(msg.format(key, value))

    data.setdefault('options', {})
    for key, value in data['options'].items():
      if isinstance(value, str):
        value = {"type": value}
      type_name = value.pop('type')
      if '.' not in type_name:
        type_name = __name__ + '._aliases.' + type_name
      try:
        option_type = pyutils.import_(type_name)
      except ImportError:
        option_type = None
      if not isinstance(option_type, type) or not issubclass(option_type, BaseOption):
        raise Manifest.Invalid('invalid option type: {!r}'.format(type_name))
      try:
        data['options'][key] = option_type(key, **value)
      except TypeError as exc:
        raise Manifest.Invalid(exc)

    try:
      return Manifest(**data)
    except TypeError as exc:
      raise Manifest.Invalid(exc)


class BaseOption(object, metaclass=abc.ABCMeta):
  """
  Base class for option value processors that convert and validate options
  from string values (usually provided via the command-line or environment
  variables).

  .. attribute:: name

  .. attribute:: inherit

  .. attribute:: help

  .. attribute:: default
  """

  def __init__(self, name, inherit=True, help=None, default=None):
    self.name = name
    self.inherit = inherit
    self.help = None
    self.default = default

  @abc.abstractmethod
  def __call__(self, value):
    """
    Convert the string *value* to the respective Python representation.
    """


class BoolOption(BaseOption):
  """
  Represents a boolean option. Supports the identifiers "yes", "true",
  "1", "no", "false" and "0".
  """

  def __init__(self, name, default=False, **kwargs):
    super().__init__(name, default=default, **kwargs)

  def __call__(self, value):
    if isinstance(value, str):
      value = value.strip().lower()
      if value in ('yes', 'true', '1'):
        return True
      elif value in ('no', 'false', '0'):
        return False
      elif value == '':
        return self.default
      else:
        raise ValueError("invalid value for bool option: {!r}".format(value))
    else:
      return bool(value)


class TripletOption(BoolOption):
  """
  Just like the :class:`BoolOption` but with a third option, accepting
  "null" and "none" (which maps to :const:`None`).
  """

  def __call__(self, value):
    try:
      return super().__call__(value)
    except ValueError:
      if isinstance(value, str):
        value = value.strip().lower()
        if value in ('null', 'none'):
          return None
      elif value is None:
        return None
    raise ValueError("invalid value for triplet option: {0!r}".format(value))


class StringOption(BaseOption):
  """
  Plain-string option.
  """

  def __init__(self, name, default='', **kwargs):
    super().__init__(name, **kwargs)
    self.default = default

  def __call__(self, string):
    return string


class PathOption(StringOption):
  """
  Option for paths. Relative paths are automatically converted to absolute
  paths. It is assumed that relative paths are specified relative to
  :attr:`Session.maindir`.
  """

  def __call__(self, value):
    from craftr.core.session import session
    if not path.isabs(value):
      value = path.join(session.maindir, value)
    return path.norm(value)


class _aliases:
  """
  This class serves as a namespace in order to be able to specify the
  standard option and loader types by their shortcut name in the manifest
  rather than by their FQN.
  """

  bool = BoolOption
  triplet = TripletOption
  string = StringOption
  path = PathOption



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.utils.tty`
=======================

This module provides colorized terminal output and other terminal helpers.
In order to colorize output, the :mod:`termcolor` module is required. On
Windows, the `colorama` module is also necessary.
"""

import os
import sys

# Attempt to import colorama and termcolor, but they are only required for
# colorization.
try:
  import colorama
except ImportError:
  colorama = None
try:
  import termcolor
except ImportError:
  termcolor = None

# Only enable colorized output if attached to a TTY or if explicitly
# requested by the environment.
isatty = (sys.stdout.isatty() and sys.stderr.isatty())
if os.environ.get('CRAFTR_ISATTY') == 'true':
  isatty = True
elif os.environ.get('CRAFTR_ISATTY') == 'false':
  isatty = False

if isatty and colorama:
  colorama.init()


def terminal_size():
  """
  Determines the size of the terminal.
  """

  if os.name == 'nt':
    # http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
    import ctypes, struct
    h = ctypes.windll.kernel32.GetStdHandle(-12)
    csbi = ctypes.create_string_buffer(22)
    res = ctypes.windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
    if res:
      (bufx, bufy, curx, cury, wattr, left, top, right,
       bottom, maxx, maxy) = struct.unpack('hhhhHhhhhhh', csbi.raw)
      sizex = right - left + 1
      sizey = bottom - top + 1
    else:
      sizex, sizey = 80, 25
    return (sizex, sizey)
  else:
    # http://stackoverflow.com/a/3010495/791713
    import fcntl, termios, struct
    h, w, hp, wp = struct.unpack('HHHH',
        fcntl.ioctl(0, termios.TIOCGWINSZ,
        struct.pack('HHHH', 0, 0, 0, 0)))
    return w, h


def clear_line():
  """
  Clears out the current line in the terminal completely and resets the
  curser to the first column.
  """

  print('\r\33[K', end='')


def colored(text, color=None, on_color=None, attrs=None):
  """
  Synonym for :func:`termcolor.colored()` that can also be used if the
  module is not available, in which case *text* is returned unchanged.
  """

  if not termcolor or not isatty:
    return text
  if isinstance(attrs, str):
    attrs = [attrs]
  elif not attrs:
    attrs = []
  return termcolor.colored(text, color, on_color, attrs)


def compile(color=None, on_color=None, attrs=None):
  """
  Compile an ANSI escape sequence and return it. Return an empty string if
  the :mod:`termcolor` module is not available. To reset the styling, use the
  :data:`reset` string.
  """

  if not termcolor or not isatty:
    return ''

  if isinstance(attrs, str):
    attrs = [attrs]
  elif not attrs:
    attrs = []

  res = ''
  fmt_str = '\033[%dm'
  for attr in attrs:
    res += fmt_str % termcolor.ATTRIBUTES[attr]
  if on_color is not None:
    res += fmt_str % termcolor.HIGHLIGHTS[on_color]
  if color is not None:
    res += fmt_str % termcolor.COLORS[color]

  return res


#: ANSI code to reset the text styling. Only available if the :mod:`termcolor`
#: module is available, otherwise it will be an empty string.
reset = termcolor.RESET if (termcolor and isatty) else ''



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import shlex
import subprocess
import sys

from . import path
from subprocess import PIPE, STDOUT

class safe(str):
  """
  If this object is passed to `quote()`, it will not be escaped.
  """

  pass


def split(s):
  """
  Enhanced implementation of :func:`shlex.split`.
  """

  result = shlex.split(s, posix=(os.name != 'nt'))
  if os.name == 'nt':
    # With posix=False, shlex.split() will not interpret \ as the
    # escape character (which is what we need on windows), but it
    # will also not eliminate quotes, which is what we need to do
    # here now.
    quotes = '\"\''
    result = [x[1:-1] if (x and x[0] in quotes and x[-1] in quotes) else x for x in result]
  return result


def quote(s, for_ninja=False):
  """
  Enhanced implementation of :func:`shlex.quote` as it generates single-quotes
  on Windows which can lead to problems.
  """

  if isinstance(s, safe):
    return s
  if os.name == 'nt' and os.sep == '\\':
    s = s.replace('"', '\\"')
    if re.search('\s', s):
      s = '"' + s + '"'
  else:
    s = shlex.quote(s)
  if for_ninja:
    # Fix escaped $ variables on Unix, see issue craftr-build/craftr#30
    s = re.sub(r"'(\$\w+)'", r'\1', s)
  return s


def format(fmt, *args, **kwargs):
  """"
  Similar to :meth:`str.format`, but this function will escape all arguments
  with the :func:`quote` function. This is useful to easily generate commands.

  .. code:: python

    >>> format('gcc {} -o {}', 'my source/main.c', 'spaces are bad/main')
    'gcc "my source/main.c" -o "spaces are bad/main"'
  """

  return fmt.format(*map(quote, args), **{k: quote(v) for k, v in kwargs.items()})


def join(cmd, for_ninja=False):
  """
  Join a list of strings to a single command string.
  """

  return ' '.join(quote(x, for_ninja=for_ninja) for x in cmd)


def find_program(name):
  """
  Finds the program *name* in the directories listed by the ``PATH``
  environment variable and returns the full absolute path to it. On Windows,
  this also takes the `PATHEXT` variable into account.

  :param name: The name of the program to find.
  :return: :class:`str` -- The absolute path to the program.
  :raise FileNotFoundError: If the program could not be found in the PATH.
  :raise PermissionError: If a candidate for "name" was found but
    it is not executable.
  """

  if path.dirname(name):
    name = path.abs(name)
  if path.isabs(name):
    if not path.isfile(name):
      raise FileNotFoundError(name)
    if not os.access(name, os.X_OK):
      raise PermissionError('{0!r} is not executable'.format(name))
    return name

  iswin = sys.platform.startswith('win32')
  iscygwin = sys.platform.startswith('cygwin')
  if iswin and '/' in name or '\\' in name:
    name = path.abs(name)
  elif iswin and path.sep in name:
    name = path.abs(name)

  if iswin:
    pathext = os.environ['PATHEXT'].split(path.pathsep)
  elif iscygwin:
    pathext = [None, '.exe']
  else:
    pathext = [None]

  first_candidate = None
  for dirname in os.environ['PATH'].split(path.pathsep):
    fullname = path.join(dirname, name)
    for ext in pathext:
      extname = (fullname + ext) if ext else fullname
      if path.isfile(extname):
        if os.access(extname, os.X_OK):
          return extname
        if first_candidate is None:
          first_candidate = extname

  if first_candidate:
    raise PermissionError('{0!r} is not executable'.format(first_candidate))
  raise FileNotFoundError(name)


def test_program(name):
  """
  Uses :func:`find_program` to find the path to *name* and returns
  True if it could be found, False otherwise.
  """

  try:
    find_program(name)
  except OSError:
    return False
  return True


class _ProcessError(Exception):
  """
  Base class that implements the attributes and behaviour of errors
  that will inherit from this exception class.
  """

  def __init__(self, process):
    self.process = process

  @property
  def returncode(self):
    return self.process.returncode

  @property
  def cmd(self):
    return self.process.cmd

  @property
  def stdout(self):
    return self.process.stdout

  @property
  def stderr(self):
    return self.process.stderr

  @property
  def output(self):
    return self.process.output


class CalledProcessError(_ProcessError):
  """
  This exception is raised when a process exits with a non-zero
  returncode and the run was to be checked for such state. The exception
  contains the process information.
  """

  def __str__(self):
    return '{0!r} exited with non-zero exit-code {1}'.format(self.cmd, self.returncode)


class TimeoutExpired(_ProcessError):
  """
  This exception is raised when a process did not exit after a
  specific timeout. If this exception was raised, the child process
  has already been killed.
  """

  def __init__(self, process, timeout):
    super().__init__(process)
    assert isinstance(timeout, (int, float))
    self.timeout = timeout

  def __str__(self):
    return '{0!r} expired timeout of {1} second(s)'.format(self.cmd, self.timeout)


class CompletedProcess(object):
  """
  This class represents a completed process.
  """

  __slots__ = 'cmd returncode stdout stderr'.split()

  def __init__(self, cmd, returncode, stdout, stderr):
    self.cmd = cmd
    self.returncode = returncode
    self.stdout = stdout
    self.stderr = stderr

  def __repr__(self):
    return '<CompletedProcess {0!r} with exit-code {1}>'.format(self.cmd, self.returncode)

  @property
  def output(self):
    return self.stdout

  @output.setter
  def output(self, value):
    self.stdout = value

  def decode(self, encoding):
    if encoding is None:
      return
    if self.stdout is not None:
      self.stdout = self.stdout.decode(encoding)
    if self.stderr is not None:
      self.stderr = self.stderr.decode(encoding)

  def check_returncode(self):
    if self.returncode != 0:
      raise CalledProcessError(self)


def run(cmd, *, stdin=None, input=None, stdout=None, stderr=None, shell=False,
    timeout=None, check=False, cwd=None, encoding=sys.getdefaultencoding()):
  """
  Run the process with the specified *cmd*. If *cmd* is a list of
  commands and *shell* is True, the list will be automatically converted
  to a properly escaped string for the shell to execute.

  .. note::

    If "shell" is True, this function will manually check if the file
    exists and is executable first and raise :class:`FileNotFoundError`
    if not.

  :raise CalledProcessError: If *check* is True and the process exited with
    a non-zero exit-code.
  :raise TimeoutExpired: If *timeout* was specified and the process did not
    finish before the timeout expires.
  :raise OSError: For some OS-level error, eg. if the program could not be
      found.
  """

  if shell and not isinstance(cmd, str):
    cmd = join(cmd)
  elif not shell and isinstance(cmd, str):
    cmd = split(cmd)

  if shell:
    # Wrapping the call in a shell invokation will not raise an
    # exception when the file could not actually be executed, thus
    # we have to check it manually.
    if isinstance(cmd, str):
      program = split(cmd)[0]
    else:
      program = cmd[0]
    find_program(program)

  try:
    popen = subprocess.Popen(
      cmd, stdin=stdin, stdout=stdout, stderr=stderr, shell=shell, cwd=cwd)
    stdout, stderr = popen.communicate(input, timeout)
  except subprocess.TimeoutExpired as exc:
    # TimeoutExpired.stderr available only since Python3.5
    stderr = getattr(exc, 'stderr', None)
    process = CompletedProcess(exc.cmd, None, exc.output, stderr)
    process.decode(encoding)
    raise TimeoutExpired(process, timeout)
  except OSError as exc:
    if not exc.filename and os.name == 'nt':
      # Windows does not include the name of the file with which
      # the error occured in the exception message.
      if isinstance(cmd, str):
        program = split(cmd)[0]
      else:
        program = cmd[0]
      exc.filename = program
    raise

  process = CompletedProcess(cmd, popen.returncode, stdout, stderr)
  process.decode(encoding)
  if check:
    process.check_returncode()
  return process


def pipe(*args, merge=True, **kwargs):
  """
  Like `run()`, but pipes stdout and stderr to a buffer instead of
  directing them to the current standard out and error files. If *merge*
  is True, stderr will be merged into stdout.
  """

  kwargs.setdefault('stdout', PIPE)
  kwargs.setdefault('stderr', STDOUT if merge else PIPE)
  return run(*args, **kwargs)



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr.utils import argspec
from craftr.utils import path
from urllib.error import URLError, HTTPError

import cgi
import urllib.request


class UserInterrupt(Exception):
  """
  An exception that represents the intended abortion of an operation.
  """


def download_file(url, filename=None, file=None, directory=None,
    on_exists='rename', progress=None, chunksize=4096, urlopen_kwargs=None):
  """
  Download a file from a URL to one of the following destinations:

  :param filename: A filename to write the downloaded file to.
  :param file: A file-like object.
  :param directory: A directory. The filename will be automatically
    determined from the ``Content-Disposition`` header received from
    the server or the last path elemnet in the URL.

  Additional parameters for the *directory* parameter:

  :param on_exists: The operation to perform when the file already exists.
    Available modes are ``rename``, ``overwrite`` and ``skip``.

  Additional parameters:

  :param progress: A callable that accepts a single parameter that is a
    dictionary with information about the progress of the download. The
    dictionary provides the keys ``size``,  ``downloaded`` and ``response``.
    If the callable returns :const:`False` (specifically the value False), the
    download will be aborted and a :class:`UserInterrupt` will be raised.
  :param urlopen_kwargs: A dictionary with additional keyword arguments
    for :func:`urllib.request.urlopen`.

  Raise and return:

  :raise HTTPError: Can be raised by :func:`urllib.request.urlopen`.
  :raise URLError: Can be raised by :func:`urllib.request.urlopen`.
  :raise UserInterrupt: If the *progress* returned :const:`False`.
  :return: If the download mode is *directory*, the name of the downloaded
    file will be returned and False if the file was newly downloaded, True
    if the download was skipped because the file already existed.
    Otherwise, the number of bytes downloaded will be returned.
  """

  argspec.validate('on_exists', on_exists, {'enum': ['rename', 'overwrite', 'skip']})

  if sum(map(bool, [filename, file, directory])) != 1:
    raise ValueError('exactly one of filename, file or directory must be specifed')

  response = urllib.request.urlopen(url, **(urlopen_kwargs or {}))

  if directory:
    try:
      filename = parse_content_disposition(
        response.headers.get('Content-Disposition', ''))
    except ValueError:
      filename = url.split('/')[-1]
    filename = path.join(directory, filename)
    path.makedirs(directory)

    if path.exists(filename):
      if on_exists == 'skip':
        return filename, True
      elif on_exists == 'rename':
        index = 0
        while True:
          new_filename = filename + '_{:0>4}'.format(index)
          if not path.exists(new_filename):
            filename = new_filename
            break
          index += 1
      elif on_exists != 'overwrite':
        raise RuntimeError

  try:
    size = int(response.headers.get('Content-Length', ''))
  except ValueError:
    size = None

  progress_info = {'response': response, 'size': size, 'downloaded': 0,
    'completed': False, 'filename': filename, 'url': url}
  if progress and progress(progress_info) is False:
    raise UserInterrupt

  def copy_to_file(fp):
    while True:
      data = response.read(chunksize)
      if not data:
        break
      progress_info['downloaded'] += len(data)
      fp.write(data)
      if progress and progress(progress_info) is False:
        raise UserInterrupt
    progress_info['completed'] = True
    if progress and progress(progress_info) is False:
      raise UserInterrupt

  if filename:
    path.makedirs(path.dirname(filename))
    try:
      with open(filename, 'wb') as fp:
        copy_to_file(fp)
    except BaseException:
      # Delete the file if it could not be downloaded successfully.
      path.remove(filename, silent=True)
      raise
  elif file:
    copy_to_file(file)

  return filename, False


def parse_content_disposition(value):
  """
  Parse the ``Content-Disposition`` header.
  """

  value, param = cgi.parse_header(value)
  if value != 'attachment':
    raise ValueError('not an attachment header')
  if 'filename' not in param:
    raise ValueError('no filename parmaeter')
  return param['filename']



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.core.build`
========================

This module provides all the API to generate a Ninja build manifest.
"""

from craftr import platform
from craftr.utils import argspec
from craftr.utils import path
from craftr.utils import pyutils
from craftr.utils import shell
from craftr.utils.singleton import Default
from ninja_syntax import Writer as NinjaWriter

import abc
import ninja_syntax
import os
import re
import stat
import sys


class DuplicateOutputError(Exception):
  """
  This exception is raised if the same output file would be built by
  multiple targets.
  """

  def __init__(self, filename, new_target, target):
    self.filename = filename
    self.new_target = new_target
    self.target = target

  def __str__(self):
    return '"{}" produced by {} and {}'.format(self.filename,
        self.new_target.name, self.target.name)


class Graph(object):
  """
  This class represents the whole build graph which is generated from
  targets reference input and output files. After a :class:`Target` is
  created, it must be added to a Graph with the :meth:`add_target` method.

  Note that all filenames in the Graph are absolute and normalized with
  the :func:`path.norm` function.

  .. attribute:: target

    Read-only. A dictionary of all the :class:`Targets<Target>` that have
    been added to the Graph.

  .. attribute:: infiles

    Read-only. A dictionary that maps normalized filenames of input files
    to a list of :class:`Targets<Target>`.

  .. attribute:: outfiles

    Read-only. A dictionary that maps normalized filenames of output files
    to the :class:`Target` that creates it.

  .. attributes:: vars

    A dictionary of variables that will be exported to the Ninja manifest.
  """

  def __init__(self):
    self.targets = {}
    self.infiles = {}
    self.outfiles = {}
    self.vars = {}
    self.tools = {}

  def add_tool(self, tool):
    """
    Add a :class:`Tool` to the Graph.

    :raise ValueError: If the :attr:`Tool.name` is already used.
    """

    if tool.name in self.tools:
      raise ValueError('a tool with the name {!r} already exists'
          .format(tool.name))
    self.tools[tool.name] = tool

  def add_target(self, target):
    """
    Add a :class:`Target` to the Graph.

    .. note:: For performance reasons, this method assumes that all paths
              in the *target* are already normalized with :meth:`path.norm`.

    :raise ValueError: If the :attr:`Target.name` is already used.
    :raise DuplicateOutputError: If the *target* lists an output file that
      is already created by another target.
    """

    argspec.validate('target', target, {'type': Target})
    if target.name in self.targets:
      raise ValueError('a target with the name {!r} already exists'
        .format(target.name))
    self.targets[target.name] = target
    for infile in target.inputs:
      self.infiles.setdefault(infile, []).append(target)
    for outfile in target.outputs:
      other = self.outfiles.setdefault(outfile, target)
      if other is not target:
        raise DuplicateOutputError(outfile, target, other)

  def export(self, writer, context, platform):
    """
    Export the build graph to a Ninja manifest.

    :param writer: A :class:`ninja_syntax.Writer` object.
    :param context: A :class:`ExportContext` object.
    :param platform: A :class:`PlatformHelper` instance.
    """

    argspec.validate('writer', writer, {"type": ninja_syntax.Writer})
    writer.comment('This file was automatically generated with Craftr.')
    writer.comment('It is not recommended to edit this file manually.')
    writer.newline()

    if self.vars:
      for key, value in self.vars.items():
        writer.variable(key, value)
      writer.newline()

    if self.tools:
      writer.comment('Tools')
      writer.comment('-----')
      for tool in self.tools.values():
        tool.export(writer, context, platform)
      writer.newline()

    defaults = []
    for target in self.targets.values():
      if not target.explicit:
        defaults.append(target.name)
      target.export(writer, context, platform)

    if defaults:
      writer.default(defaults)

class Target(object):
  """
  A higher level abstraction of a Target that can be added to a :class:`Graph`
  and then exported into a Ninja build manifest. A target should be treated
  as read-only always.
  """

  def __init__(self, name, commands, inputs, outputs, implicit_deps=(),
               order_only_deps=(), pool=None, deps=None, depfile=None,
               msvc_deps_prefix=None, explicit=False, foreach=False,
               description=None, metadata=None, cwd=None, environ=None,
               frameworks=()):
    argspec.validate('name', name, {'type': str})
    argspec.validate('commands', commands,
      {'type': list, 'allowEmpty': False, 'items':
        {'type': list, 'allowEmpty': False, 'items': {'type': [Tool, Target, str]}}})
    argspec.validate('inputs', inputs, {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('outputs', outputs, {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('implicit_deps', implicit_deps, {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('order_only_deps', order_only_deps, {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('pool', pool, {'type': [None, str]})
    argspec.validate('deps', deps, {'type': [None, str], 'enum': ['msvc', 'gcc']})
    argspec.validate('depfile', depfile, {'type': [None, str]})
    argspec.validate('msvc_deps_prefix', msvc_deps_prefix, {'type': [None, str]})
    argspec.validate('explicit', explicit, {'type': bool})
    argspec.validate('foreach', foreach, {'type': bool})
    argspec.validate('description', description, {'type': [None, str]})
    argspec.validate('metadata', metadata, {'type': [None, dict]})
    argspec.validate('cwd', cwd, {'type': [None, str]})
    argspec.validate('environ', environ, {'type': [None, dict]})
    argspec.validate('frameworks', frameworks, {'type': [list, tuple], 'items': {'type': dict}})

    # Make sure we have a copy of the implicit_deps so we can modify
    # it safely.
    implicit_deps = list(implicit_deps)

    # Expand Target objects listed in the commands.
    new_commands = []
    for command in commands:
      new_command = []
      for index, arg in enumerate(command):
        if isinstance(arg, Target):
          if index == 0 and len(arg.outputs) != 1:
            raise ValueError('Target as program in commands list must have '
                'exactly one output, has {} (target: {})'.format(
                  len(arg.outputs), arg.name))
          new_command += arg.outputs
          implicit_deps += arg.outputs
        else:
          new_command.append(arg)
      new_commands.append(new_command)

    self.name = name
    self.commands = new_commands
    self.inputs = list(map(path.norm, inputs))
    self.outputs = list(map(path.norm, outputs))
    self.implicit_deps = list(map(path.norm, implicit_deps))
    self.order_only_deps = list(map(path.norm, order_only_deps))
    self.pool = pool
    self.deps = deps
    self.depfile = depfile
    self.msvc_deps_prefix = msvc_deps_prefix
    self.explicit = explicit
    self.foreach = foreach
    self.description = description
    self.metadata = metadata or {}
    self.cwd = cwd
    self.environ = environ or {}
    self.frameworks = frameworks

    if self.foreach and len(self.inputs) != len(self.outputs):
      raise ValueError('foreach target must have the same number of output '
        'files ({}) as input files ({})'.format(len(self.outputs),
        len(self.inputs)))

    if self.deps == 'gcc' and not self.depfile:
      raise ValueError('require depfile with deps="gcc"')

  def __str__(self):
    return '<{}.Target "{}">'.format(__name__, self.name)

  def __lshift__(self, other):
    """
    Adds *other* as an implicit dependency to the target.

    :param other: A :class:`Target` or :class:`str`.
    :return: ``self``
    """

    if isinstance(other, Target):
      self.implicit_deps += other.outputs
    elif isinstance(other, str):
      self.implicit_deps.append(path.norm(other))
    else:
      raise TypeError("Target.__lshift__() expected Target or str")
    return self

  def export(self, writer, context, platform):
    """
    Export the target to a Ninja manifest.
    """

    writer.comment("target: {}".format(self.name))
    writer.comment("--------" + "-" * len(self.name))
    commands = platform.prepare_commands([list(map(str, c)) for c in self.commands])

    # Check if we need to export a command file or can export the command
    # directly.
    if not self.environ and len(commands) == 1:
      commands = [platform.prepare_single_command(commands[0], self.cwd)]
    else:
      filename = path.join('.commands', self.name)
      command, __ = platform.write_command_file(filename, commands,
        self.inputs, self.outputs, cwd=self.cwd, environ=self.environ,
        foreach=self.foreach)
      commands = [command]

    assert len(commands) == 1
    command = shell.join(commands[0], for_ninja=True)

    writer.rule(self.name, command, pool=self.pool, deps=self.deps,
      depfile=self.depfile, description=self.description)

    if self.msvc_deps_prefix:
      # We can not write msvc_deps_prefix on the rule level with Ninja
      # versions older than 1.7.1. Write it global instead, but that *could*
      # lead to issues...
      indent = 1 if context.ninja_version > '1.7.1' else 0
      writer.variable('msvc_deps_prefix', self.msvc_deps_prefix, indent)

    writer.newline()
    if self.foreach:
      assert len(self.inputs) == len(self.outputs)
      for infile, outfile in zip(self.inputs, self.outputs):
        writer.build(
          [outfile],
          self.name,
          [infile],
          implicit=self.implicit_deps,
          order_only=self.order_only_deps)
    else:
      writer.build(
        self.outputs or [self.name],
        self.name,
        self.inputs,
        implicit=self.implicit_deps,
        order_only=self.order_only_deps)

    if self.outputs and self.name not in self.outputs and not self.explicit:
      writer.build(self.name, 'phony', self.outputs)


class Tool(object):
  """
  This class represents a program that can be called by by the command in
  build rules. Usually this is only necessary if the tool requires a special
  environment.

  Depending on whether its necessary on the current platform, tools might be
  exported into shell scripts. The actual command to be called in the command
  can then be retrieved by converting the :class:`Tool` object to a string.

  .. attribute:: name

    The name of the tool. In a build :class:`Graph`, there may only be one tool
    associated with the same name.

  .. attribute:: command

    The command that is to be executed and to which the additional command-line
    arguments are forwarded to. Must be a list of strings.

  .. attribute:: preamble

    A list of commands/files that are to be sourced before the :attr:`command`
    can be executed.

  .. attribute:: environ

    A dictionary of environment variables that will be exported before the
    command or premable is executed.

  .. attribute:: exported_command

    After :meth:`exported_command` has been called, this member contains the
    command that was exported for the Tool. This is a single string.
  """

  def __init__(self, name, command, preamble=None, environ=None):
    argspec.validate('name', name, {'type': str})
    argspec.validate('command', command,
        {'type': [list, tuple], 'allowEmpty': False, 'items': {'type': str}})
    argspec.validate('preamble', preamble,
        {'type': [None, list, tuple], 'items':
          {'type': [list, tuple], 'allowEmpty': False, 'items': {'type': str}}})
    argspec.validate('environ', environ, {'type': [None, dict]})

    self.name = name
    self.command = command
    self.preamble = preamble or []
    self.environ = environ or {}
    self.exported_command = None

  def __str__(self):
    """
    Returns the name of the command to execute the tool in Ninja. This is
    always a variable reference.
    """

    return '$CraftrTool_{}'.format(self.name.replace('.', '_'))

  def export(self, writer, context, platform):
    name = str(self)[1:]
    if not self.preamble and not self.environ:
      self.exported_command = shlex.join(self.command)
    else:
      filename = path.join('.tools', name)
      command, filename = platform.write_command_file(
          filename, list(self.preamble) + [self.command], environ=self.environ,
          accept_additional_args=True)
      self.exported_command = shell.join(command)
    writer.variable(name, self.exported_command)


class ExportContext(object):
  """
  An instance of this class is required for :meth:`Graph.export` and
  :meth:`Target.export`. It provides additional information that can influence
  the exported manifest.

  .. attribute:: ninja_version
  """

  def __init__(self, ninja_version):
    self.ninja_version = ninja_version


class PlatformHelper(object, metaclass=abc.ABCMeta):
  """
  Interface to abstract platfrom dependent operations during the export
  process.
  """

  @abc.abstractmethod
  def prepare_commands(self, commands):
    """
    Passed a list of list of strings that represents the commands for
    a target. Return the same value or a modified value.
    """

  @abc.abstractmethod
  def prepare_single_command(self, command, cwd):
    """
    Given a single command as a list of strings and an optional working
    directory path, return an updated list of strings that serves as the
    new command including the current working directory switch.
    """

  @abc.abstractmethod
  def write_command_file(self, filename, commands, inputs=None, outputs=None,
      cwd=None, environ=None, foreach=False, suffix=Default, dry=False,
      accept_additional_args=False):
    """
    Writes a file that can be run by the native shell to execute the *commands*.
    If *suffix* is omitted, the default script suffix for the shell will be
    appended (which is especially important on Windows).

    References to ``$in`` and ``$out`` will be replaced by the values of
    *inputs* and *outputs* if they are specified. Note that in a concatenated
    context (eg. ``$out.d``) the number input or output files (respectively)
    must be exactly one, otherwise an error is raised.

    If *foreach* is specified, the generated command file accepts two
    command-line arguments for the input and output file. Note that references
    to ``$in`` and ``$out`` will then be replaced by the respective placeholder
    in the shells scripting language and the *inputs* and *outputs* parmaeters
    are ignored.

    If *accept_additional_args* is True, the last command in *commands* must
    accept forward arguments passed to the shell script.

    Returns 1) a list of strings that represents the command to execute the
    script. If *foreach* is specified, the two last items of the returned list
    will be ``['$in', '$out']``. 2) The actual filename that has been created.
    """

  @staticmethod
  def replace_argument_inout_vars(arg, inputs, outputs):
    """
    Helper function for :meth:`write_command_file` to replace the ``$in``
    and ``$out`` variables in a command argument *arg* with the specified
    *inputs* and *outputs*.
    """

    def replace_var(var, value, files):
      files = files or ()
      match = re.match('.*(\\$' + var + ')\\b.*', value)
      if not match:
        return [value]
      if match.group(0) == '$' + var:
        return files
      if len(files) != 1:
        raise RuntimeError('embedded in an argument, `$' + var +
          '` can only expand with one filename')
      return [value[:match.start(1)] + files[0] + value[match.end(1):]]

    result = replace_var('in', arg, inputs)
    if len(result) == 1:
      result = replace_var('out', result[0], outputs)
    return result

  @staticmethod
  def replace_commands_inout_vars(commands, inputs, outputs):
    """
    Applies :meth:`replace_argument_inout_vars` to all *commands*.
    """

    new_commands = []
    for command in commands:
      new_commands.append(pyutils.flatten(
        [PlatformHelper.replace_argument_inout_vars(x, inputs, outputs)
        for x in command]))
    return new_commands
    commands = new_commands


class WindowsPlatformHelper(PlatformHelper):

  def prepare_commands(self, commands):
    # ISSUE craftr-build/craftr#67
    # On Windows, commands that are not executables need to be invoked via CMD.
    new_commands = []
    for args in commands:
      try:
        prog = shell.find_program(args[0])
        is_executable = prog.endswith('.exe')
      except FileNotFoundError:
        is_executable = args[0].endswith('.exe')
      if not is_executable:
        args = ['cmd', '/c'] + args
      new_commands.append(args)
    return new_commands

  def prepare_single_command(self, command, cwd):
    if cwd is not None:
      command = ['cmd', '/c', 'cd', cwd, shell.safe('&&')] + command
    return command

  def write_command_file(self, filename, commands, inputs=None, outputs=None,
      cwd=None, environ=None, foreach=False, suffix='.cmd', dry=False,
      accept_additional_args=False):

    if suffix is not None:
      filename = path.addsuffix(filename, suffix)
    result = ['cmd', '/Q', '/c', filename]
    if foreach:
      result += ['$in', '$out']
      inputs, outputs = ['%1'], ['%2']

    commands = self.replace_commands_inout_vars(commands, inputs, outputs)
    if dry:
      return result, filename

    path.makedirs(path.dirname(path.abs(filename)))
    with open(filename, 'w') as fp:
      fp.write('REM This file is automatically generated with Craftr. It is \n')
      fp.write('REM not recommended to modify it manually.\n\n')
      if cwd is not None:
        fp.write('cd ' + shell.quote(cwd) + '\n\n')
      for key, value in environ.items():
        fp.write('set ' + shell.quote('{}={}'.format(key, value), for_ninja=True) + '\n')
      fp.write('\n')
      for index, command in enumerate(commands):
        if accept_additional_args and index == len(commands)-1:
          command.append(shell.safe('%*'))
        fp.write(shell.join(command) + '\n')
        fp.write('if %errorlevel% neq 0 exit %errorlevel%\n\n')

    return result, filename


class UnixPlatformHelper(PlatformHelper):

  def prepare_commands(self, commands):
    return commands

  def prepare_single_command(self, command, cwd):
    if cwd is not None:
      command = [shell.safe('('), 'cd', cwd, shell.safe('&&')] + command + [shell.safe(')')]
    return command

  def write_command_file(self, filename, commands, inputs=None, outputs=None,
      cwd=None, environ=None, foreach=False, suffix='.sh', dry=False,
      accept_additional_args=False):

    if suffix is not None:
      filename = path.addsuffix(filename, suffix)
    result = [filename]
    if foreach:
      result += ['$in', '$out']
      inputs, outputs = ['%1'], ['%2']

    commands = self.replace_commands_inout_vars(commands, inputs, outputs)
    if dry:
      return result, filename

    path.makedirs(path.dirname(filename))
    with open(filename, 'w') as fp:
      # TODO: Make sure this also works for shells other than bash.
      fp.write('#!' + shell.find_program(environ.get('SHELL', 'bash')) + '\n')
      fp.write('set -e\n')
      if cwd:
        fp.write('cd ' + shell.quote(cwd) + '\n')
      fp.write('\n')
      for key, value in environ.items():
        fp.write('export {}={}\n'.format(key, shell.quote(value)))
      fp.write('\n')
      for index, command in enumerate(commands):
        if accept_additional_args and index == len(commands)-1:
          command.append(shell.safe('$*'))
        fp.write(shell.join(command))
        fp.write('\n')

    os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
      stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH)  # rwxrw-r--

    return result, filename


def get_platform_helper():
  if platform.name == 'win':
    return WindowsPlatformHelper()
  elif platform.name in ('cygwin', 'linux', 'mac'):
    return UnixPlatformHelper()
  raise EnvironmentError('unsupported platform: {}'.format(platform.name))



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

options = session.module.options
from craftr.loaders import external_archive

# Framework that is used by other libraries/applications.
cURL = Framework(
  include = [],
  defines = ['CURL_STATICLIB'] if options.static else [],
  libs = []
)

# Framework for building libcURL.
cURL_building = Framework(
  frameworks = [cURL],
  defines = ['BUILDING_LIBCURL'],
  include = []
)

# Platform dependent settings.
if platform.name == 'win':
  cURL['libs'] += 'kernel32 user32 gdi32 winspool shell32 ole32 oleaut32 uuid comdlg32 advapi32 wldap32 winmm ws2_32 crypt32'.split()
  cURL_building['defines'] += ['HAVE_CONFIG_H', '_WIN32_WINNT=0x0501']
  # Note: Yes, a bit hacky. Workaround until we can either maybe evaluate
  # CMake files or have CMake like feature checks (see craftr-build/craftr#134).
  cURL_building['include'] += [local('windows')]
else:
  # TODO
  error('platform currently not supported: {}'.format(platform.name))

# Grab the cURL source and update the include directory in the public framework.
source_directory = external_archive(
  "https://curl.haxx.se/download/curl-{}.tar.gz".format(options.version)
)
cURL['include'] += [path.join(source_directory, 'include')]

# Compile the library.
load_module('craftr.lang.cxx.*')
libcURL = cxx_library(
  link_style = 'static' if options.static else 'shared',
  inputs = c_compile(
    sources = glob(['src/**/*.c', 'lib/**/*.c'], parent = source_directory),
    include = [path.join(source_directory, 'lib')],
    frameworks = [cURL, cURL_building]
  ),
  output = 'cURL'
)

cxx_extend_framework(cURL, libcURL)



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.targetbuilder`
===========================

Provides the :class:`TargetBuilder` and :class:`Framework` classes.
"""

from craftr.core import build
from craftr.core.logging import logger
from craftr.core.session import session
from craftr.utils import argspec, pyutils
from nr.py.bytecode import get_assigned_name

import collections
import sys


def get_full_name(target_name, module=None):
  """
  Given a *target_name*, this function generates the fully qualified name of
  that target that includes the *module* name and version. If *module* is
  omitted, the currently executed module is used.
  """

  if module is None:
    if not session:
      raise RuntimeError('no session context')
    module = session.module
    if not module:
      raise RuntimeError('no current module')

  return '{}.{}'.format(module.ident, target_name)


def gtn(target_name=None, name_hint=NotImplemented):
  """
  This function is mandatory in combination with the :class:`TargetBuilder`
  class to ensure that the correct target name can be deduced. If the
  *target_name* parmaeter is specified, it is returned right away. Otherwise,
  if it is :const:`None`, :func:`get_assigned_name` will be used to retrieve
  the variable name of the expression of 2 frames above this call.

  To visualize the behaviour, imagine you are using a target generator function
  like this:

  .. code:: python

      objects = cxx_compile(
        srcs = glob(['src/*.cpp'])
      )

  The ``cxx_compile()`` function would use the :class:`TargetBuilder`
  similar to what is shown below.

  .. code:: python

      builder = TargetBuilder(inputs, frameworks, kwargs, name = gtn(name))

  And the :func:`gtn` function will use the frame that ``cxx_compile()`` was
  called from to retrieve the name of the variable "objects". In the case that
  an explicit target name is specified, it will gracefully return that name
  instead of figuring the name of the assigned variable.

  If *name_hint* is :const:`None` and no assigned name could be determined,
  no exception will be raised but also no alternative target name will
  be generated and :const:`None` will be returned. This is useful for wrapping
  existing target generator functions.
  """

  if not session:
    raise RuntimeError('no session context')
  module = session.module
  if not module:
    raise RuntimeError('no current module')

  if target_name is None:
    try:
      target_name = get_assigned_name(sys._getframe(2))
    except ValueError:
      if name_hint is NotImplemented:
        raise
  elif '-' in target_name:
    # TODO: We should find a better way to determine if the target is
    # TODO: already an absolute target name.
    return target_name

  full_name = None
  if target_name is None:
    if name_hint is None:
      return None

    index = 0
    while True:
      target_name = '{}_{:0>4}'.format(name_hint, index)
      full_name = get_full_name(target_name, module)
      if full_name not in session.graph.targets:
        break
      index += 1

  if full_name is None:
    full_name = get_full_name(target_name, module)

  return full_name


class TargetBuilder(object):
  """
  This is a helper class for target generators that does a lot of convenience
  handling such as creating a :class:`OptionMerge` from the build options
  specified with the *frameworks* argument or from the input
  :class:`Targets<build.Target>`.

  :param name: The name of the target. Derive this target name by using
    the :func:`gtn` function.
  :param option_kwargs: A dictionary of additional call-level options
    that have been passed to the target generator function. These will
    take top priority in the :class:`OptionMerge`.
  :param inputs: A list of input files or :class:`build.Target` objects
    from which the outputs will be used as input files and the build
    options will be included in the :class:`OptionMerge`.
  :param frameworks: A list of build :class:`Framework` objects that will be
    included in the :class:`OptionMerge`.
  :param outputs: A list of output filenames.
  :param implicit_deps: A list of filenames added as implicit dependencies.
  :param order_only_deps: A list of filenames added as order only dependencies.
  """

  def __init__(self, name, option_kwargs=None, frameworks=(), inputs=(),
      outputs=(), implicit_deps=(), order_only_deps=()):
    argspec.validate('name', name, {'type': str})
    argspec.validate('option_kwargs', option_kwargs,
        {'type': [None, dict, Framework]})
    argspec.validate('frameworks', frameworks,
        {'type': [list, tuple], 'items': {'type': Framework}})
    argspec.validate('inputs', inputs,
        {'type': [list, tuple, build.Target],
          'items': {'type': [str, build.Target]}})
    argspec.validate('outputs', outputs,
        {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('implicit_deps', implicit_deps,
        {'type': [list, tuple], 'items': {'type': str}})
    argspec.validate('order_only_deps', order_only_deps,
        {'type': [list, tuple], 'items': {'type': str}})

    self.frameworks = list(frameworks)
    if isinstance(inputs, build.Target):
      inputs = [inputs]

    # If we find any Target objects in the inputs, expand the outputs
    # and append the frameworks.
    if inputs is not None:
      self.inputs = []
      for input_ in (inputs or ()):
        if isinstance(input_, build.Target):
          pyutils.unique_extend(self.frameworks, input_.frameworks)
          self.inputs += input_.outputs
        else:
          self.inputs.append(input_)
    else:
      self.inputs = None

    if option_kwargs is None:
      option_kwargs = {}

    self.name = name
    self.option_kwargs = Framework(name, **option_kwargs)
    self.option_kwargs_defaults = Framework(name + "_defaults")
    self.options_merge = OptionMerge(self.option_kwargs,
        self.option_kwargs_defaults, *self.frameworks)
    assert self.option_kwargs in self.options_merge.frameworks
    self.outputs = list(outputs)
    self.implicit_deps = list(implicit_deps)
    self.order_only_deps = list(order_only_deps)
    self.metadata = {}
    self.used_option_keys = set()

  def get(self, key, default=None):
    self.used_option_keys.add(key)
    return self.options_merge.get(key, default)

  def get_list(self, key):
    self.used_option_keys.add(key)
    return self.options_merge.get_list(key)

  def add_local_framework(self, *args, **kwargs):
    fw = Framework(*args, **kwargs)
    self.options_merge.append(fw)

  def setdefault(self, key, value):
    self.option_kwargs_defaults[key] = value

  def build(self, commands, inputs=(), outputs=(), implicit_deps=(),
      order_only_deps=(), metadata=None, **kwargs):
    """
    Create a :class:`build.Target` from the information in the builder,
    add it to the build graph and return it.
    """

    unused_keys = set(self.option_kwargs.keys()) - self.used_option_keys
    if unused_keys:
      logger.warn('TargetBuilder: "{}" unhandled option keys'.format(self.name))
      with logger.indent():
        for key in unused_keys:
          logger.warn('[-] {}={!r}'.format(key, self.option_kwargs[key]))

    # TODO: We could make this a bit shorter..
    inputs = self.inputs + list(inputs or ())
    outputs = self.outputs + list(outputs or ())
    implicit_deps = self.implicit_deps + list(implicit_deps or ())
    order_only_deps = self.order_only_deps + list(order_only_deps or ())
    if metadata is None:
      metadata = self.metadata
    elif self.metadata:
      raise RuntimeError('metadata specified in constructor and build()')

    implicit_deps = list(implicit_deps)
    for item in self.get_list('implicit_deps'):
      if isinstance(item, build.Target):
        implicit_deps += item.outputs
      elif isinstance(item, str):
        implicit_deps.append(item)
      else:
        raise TypeError('expected Target or str in "implicit_deps", found {}'
            .format(type(item).__name__))

    target = build.Target(self.name, commands, inputs, outputs, implicit_deps,
        order_only_deps, metadata=metadata, frameworks=self.frameworks, **kwargs)
    session.graph.add_target(target)
    return target


class Framework(dict):
  """
  A framework is simply a dictionary with a name to identify it. Frameworks
  are used to represent build options.
  """

  def __init__(self, __name=None, **kwargs):
    super().__init__(**kwargs)
    self.name = gtn(__name)

  def __repr__(self):
    return '<Framework "{}": {}>'.format(self.name, super().__repr__())


class OptionMerge(object):
  """
  This class represents a virtual merge of :class:`Framework` objects. Keys
  in the first dictionaries passed to the constructor take precedence over the
  last.

  :param frameworks: One or more :class:`Framework` objects. Note that
    the constructor will expand and flatten the ``'frameworks'`` list.
  """

  def __init__(self, *frameworks):
    self.frameworks = []
    [self.append(x) for x in frameworks]

  def __getitem__(self, key):
    for options in self.frameworks:
      try:
        return options[key]
      except KeyError:
        pass  # intentional
    raise KeyError(key)

  def append(self, framework):
    def update(fw):
      if not isinstance(framework, Framework):
        raise TypeError('expected Framework, got {}'.format(type(fw).__name__))
      pyutils.unique_append(self.frameworks, fw, id_compare=True)
      [update(x) for x in fw.get('frameworks', [])]
    update(framework)

  def get(self, key, default=None):
    try:
      return self[key]
    except KeyError:
      return default

  def get_list(self, key):
    """
    This function returns a concatenation of all list values saved under the
    specified *key* in all option dictionaries in this OptionMerge object.
    It gives an error if one option dictionary contains a non-sequence for
    *key*.
    """

    result = []
    for option in self.frameworks:
      value = option.get(key)
      if value is None:
        continue
      if not isinstance(value, collections.Sequence):
        raise ValueError('found "{}" for key "{}" which is a non-sequence'
            .format(tpye(value).__name__, key))
      result += value
    return result



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

def make_singleton(name, type_name=None, as_bool=True):
  """
  Create a single type and return its only instance with the
  specified #name. If #type_name is not specified, it is automatically
  derived from the singleton #name.
  """

  class singleton_class(object):
    __instance = None
    def __new__(cls):
      if cls.__instance is None:
        cls.__instance = super().__new__(cls)
      return cls.__instance
    def __str__(self):
      return name
    def __repr__(self):
      return name
    def __bool__(self):
      return as_bool
    __nonzero__ = __bool__

  if type_name is None:
    type_name = name + 'Type'
  singleton_class.__name__ = type_name
  return singleton_class()

#: A singleton object like :const:`None`, :const:`True` or :const:`False`.
#: Can be used to denote the default value of an argument when :const:`None`
#: is an accepted non-default value.
Default = make_singleton('Default')



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

def primes(int kmax):
    cdef int n, k, i
    cdef int p[1000]
    result = []
    if kmax > 1000:
        kmax = 1000
    k = 0
    n = 2
    while k < kmax:
        i = 0
        while i < k and n % p[i] != 0:
            i = i + 1
        if i == k:
            p[k] = n
            k = k + 1
            result.append(n)
        n = n + 1
    return result



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.defaults`
======================

This module provides the default global namespace for Craftr modules. Names
starting with an underscore will be ignored.
"""

from craftr.core import build as _build
from craftr.core.logging import logger
from craftr.core.manifest import Namespace
from craftr.core.session import session, ModuleNotFound
from craftr.utils import path, shell
from craftr.targetbuilder import gtn, TargetBuilder, Framework
from craftr import platform

import builtins as _builtins
import itertools as _itertools
import os as _os
import require
import sys as _sys

require = require.Require(write_bytecode=False)


class ToolDetectionError(Exception):
  pass


class ModuleError(Exception):
  pass


def include_defs(filename, globals=None):
  """
  Uses :mod:`require` to load a Python file and then copies all symbols
  that do not start with an underscore into the *globals* dictionary. If
  *globals* is not specified, it will fall back to the globals of the frame
  that calls the function.
  """

  module = require(filename, _stackdepth=1)
  if globals is None:
    globals = _sys._getframe(1).f_globals
  for key, value in vars(module).items():
    if not key.startswith('_'):
      globals[key] = value


def glob(patterns, parent=None, exclude=(), include_dotfiles=False):
  """
  Wrapper for :func:`path.glob` that automatically uses the current modules
  project directory for the *parent* argument if it has not been specifically
  set.
  """

  if parent is None and session and session.module:
    parent = session.module.project_dir

  return path.glob(patterns, parent, exclude, include_dotfiles)


def local(rel_path):
  """
  Given a relative path, returns the absolute path relative to the current
  module's project directory.
  """

  parent = session.module.project_dir
  return path.norm(rel_path, parent)


def buildlocal(rel_path):
  """
  Given a relative path, returns the path (still relative) to the build
  directory for the current module. This is basically a shorthand for
  prepending the module name and version to *path*.
  """

  if path.isabs(rel_path):
    return rel_path
  return path.canonical(path.join(session.module.ident, rel_path))


def relocate_files(files, outdir, suffix, replace_suffix=True, parent=None):
  """
  Converts a list of filenames, relocating them to *outdir* and replacing
  their existing suffix. If *suffix* is a callable, it will be passed the
  new filename and expected to return the same filename, eventually with
  a different suffix.
  """

  if parent is None:
    parent = session.module.project_dir
  result = []
  for filename in files:
    filename = path.join(outdir, path.rel(filename, parent))
    filename = path.addsuffix(filename, suffix, replace=replace_suffix)
    result.append(filename)
  return result


def filter(predicate, iterable):
  """
  Alternative for the built-in ``filter()`` function that returns a list
  instead of an iterable (which is the behaviour since Python 3).
  """

  result = []
  for item in iterable:
    if predicate(item):
      result.append(item)
  return result


def map(procedure, iterable):
  """
  Alternative for the built-in ``map()`` function that returns a list instead
  of an iterable (which is the behaviour since Python 3).
  """

  result = []
  for item in iterable:
    result.append(procedure(item))
  return result


def zip(*iterables, fill=NotImplemented):
  """
  Alternative to the Python built-in ``zip()`` function. This function returns
  a list rather than an iterable and also supports swapping to the
  :func:`itertools.izip_longest` version if the *fill* parameter is specified.
  """

  if fill is NotImplemented:
    return list(_builtins.zip(*iterables))
  else:
    return list(_itertools.zip_longest(*iterables, fillvalue=fill))


def load_module(name, into=None, get_namespace=True, _stackframe=1):
  """
  Load a Craftr module by name and return it. If *into* is specified, it must
  be a dictionary that will be filled with all the members of the module. Note
  that this function returns the namespace object of the module rather than
  the actual :class:`craftr.core.session.Module` object that wraps the module
  information unless *get_namespace* is False.

  The version criteria is read from the current module's manifest.

  :param name: The name of the module to load. If this name is suffixed
    with the two characters ``.*`` and the *into* parameter is :const:`None`,
    the contents of the module will be exported into the globals of the
    calling frame.
  :param into: If specified, must be a dictionary.
  :param get_namespace:
  :return: The module namespace object (of type :class:`types.ModuleType`)
    or the actual :class:`craftr.core.session.Module` if *get_namespace*
    is False.
  :raise ModuleNotFound: If the module could not be found.
  :raise RuntimeError: If the module that is attempted to be loaded is not
    declared in the current module's manifest.

  Examples:

  .. code:: python

    cxx = load_module('craftr.lang.cxx')
    load_module('craftr.lang.cxx.*')
    assert cxx.c_compile is c_compile
  """

  if name.endswith('.*') and into is None:
    name = name[:-2]
    into = _sys._getframe(_stackframe).f_globals

  if not session:
    raise RuntimeError('no session context')
  module = session.module
  if not module:
    raise RuntimeError('no current module')

  if name not in module.manifest.dependencies:
    raise RuntimeError('"{}" can not load "{}", make sure that it is listed '
        'in the dependencies'.format(module.ident, name))

  loaded_module = session.find_module(name, module.manifest.dependencies[name])
  if not loaded_module.executed:
    loaded_module.run()

  if into is not None:
    module_builtins = frozenset('loader project_dir options'.split())
    all_vars = getattr(loaded_module.namespace, '__all__', None)
    for key, value in vars(loaded_module.namespace).items():
      if all_vars is not None:
        if key in all_vars:
          into[key] = value
      else:
        if not key.startswith('_') and key not in module_builtins and key not in globals():
          into[key] = value

  if get_namespace:
    return loaded_module.namespace
  return loaded_module


def load_file(filename, export_default_namespace=True):
  """
  Loads a Python file into a new module-like object and returns it. The
  *filename* is assumed relative to the currently executed module's
  directory (NOT the project directory which can be different).
  """

  module = session.module
  __name__ = module.ident + ':' + filename
  if not path.isabs(filename):
    filename = path.join(module.directory, filename)
  filename = path.norm(filename)

  with open(filename, 'r') as fp:
    code = compile(fp.read(), filename, 'exec')

  scope = Namespace()
  if export_default_namespace:
    vars(scope).update(module.get_init_globals())
    scope.__module__ = module.namespace
  scope.__file__ = filename
  scope.__name__ = __name__
  exec(code, vars(scope))

  return scope


def gentool(commands, preamble=None, environ=None, name=None):
  """
  Create a :class:`~_build.Tool` object. The name of the tool will be derived
  from the variable name it is assigned to unless *name* is specified.
  """

  tool = _build.Tool(gtn(name), commands, preamble, environ)
  session.graph.add_tool(tool)
  return tool


def gentarget(commands, inputs=(), outputs=(), *args, **kwargs):
  """
  Create a :class:`~_build.Target` object. The name of the target will be
  derived from the variable name it is assigned to unless *name* is specified.
  """

  target = _build.Target(gtn(kwargs.pop('name', None)), commands, inputs,
      outputs, *args, **kwargs)
  session.graph.add_target(target)
  return target


def runtarget(target, *args, inputs=(), outputs=(), **kwargs):
  """
  Simplification of :func:`gentarget` to make it more obvious that a
  generate target is actually executed.
  """

  name = gtn(kwargs.pop('name', None))
  kwargs.setdefault('explicit', True)
  kwargs.setdefault('pool', 'console')
  return gentarget([[target] + list(args)], inputs, outputs, name=name, **kwargs)


def write_response_file(arguments, builder=None, name=None, force_file=False):
  """
  Creates a response-file with the specified *name* in the in the
  ``buildfiles/`` directory and writes the *arguments* list quoted into
  the file. If *builder* is specified, it must be a :class:`TargetBuilder`
  and the response file will be added to the implicit dependencies.

  If *force_file* is set to True, a file will always be written. Otherwise,
  the function will into possible limitations of the platform and decide
  whether to write a response file or to return the *arguments* as is.

  Returns a tuple of ``(filename, arguments)``. If a response file is written,
  the returned *arguments* will be a list with a single string that is the
  filename prepended with ``@``. The *filename* part can be None if no
  response file needed to be exported.
  """

  if not name:
    if not builder:
      raise ValueError('builder must be specified if name is bot')
    name = builder.name + '.response.txt'

  if platform.name != 'win':
    return None, arguments

  # We'll just assume that there won't be more than 2048 characters for
  # other flags. The windows max buffer size is 8192.
  content = shell.join(arguments)
  if len(content) < 6144:
    return None, arguments

  filename = buildlocal(path.join('buildfiles', name))
  if builder:
    builder.implicit_deps.append(filename)

  if session.builddir:
    path.makedirs(path.dirname(filename))
    with open(filename, 'w') as fp:
      fp.write(content)
  return filename, ['@' + filename]


def error(*message):
  """
  Raises a :class:`ModuleError`.
  """

  raise ModuleError(' '.join(map(str, message)))


def append_PATH(*paths):
  """
  This is a helper function that is used to generate a ``PATH`` environment
  variable from the value that already exists and add the specified *paths*
  to it. It is typically used for example like this:

  .. code:: python

    run = gentarget(
      commands = [[main, local('example.ini')]],
      explicit=True,
      environ = {'PATH': append_PATH(qt5.bin_dir if qt5 else None)}
    )
  """

  result = _os.getenv('PATH')
  paths = _os.path.pathsep.join(filter(bool, paths))
  if paths:
    result += _os.path.pathsep + paths
  return result



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import sys


def flatten(iterable):
  """
  Flattens two levels of nested iterables into a single list.
  """

  return [item for subiterable in iterable for item in subiterable]


def import_(fqn):
  """
  Given a fully qualified name, imports the object and returns it.

  :param fqn: The full name of the object to import, including the module
    name to import it from.
  :raise ImportError: If the object can not be imported.
  :return: any
  """

  parts = iter(fqn.split('.'))
  snake = ''

  # Import module and submodules until we can no longer import modules.
  result = None
  imp_error = None
  for part in parts:
    try:
      # fromlist argument non-empty to get bottom-most module returned.
      result = __import__(snake + part, fromlist=['foo'], level=0)
    except ImportError as exc:
      imp_error = exc
      break
    snake += part + '.'
    part = None

  if result is None and imp_error:
    raise imp_error
  elif result is None:
    raise ImportError(snake.rstrip('.'))

  try:
    if part:
      result = getattr(result, part)
    for part in parts:
      result = getattr(result, part)
  except AttributeError:
    raise ImportError(fqn)

  return result


def unique_append(lst, item, id_compare=False):
  if id_compare:
    for x in lst:
      if x is item:
        return
  else:
    if item in lst:
      return
  lst.append(item)


def unique_extend(lst, iterable, id_compare=False):
  for item in iterable:
    unique_append(lst, item, id_compare)


def unique_list(iterable, id_compare=False):
  result = []
  for item in iterable:
    unique_append(result, item)
  return result


def strip_flags(command, flags):
  """
  Remove all occurences of all *flags* in the *command* list. The list is
  mutated directlry.

  :param command: A list of command-line arguments.
  :param remove_flags: An iterable of flags to remove.
  :return: The *command* parameter.
  """

  # Remove the specified flags and keep every flag that could not
  # be removed from the command.
  flags = set(flags)
  for flag in list(flags):
    count = 0
    while True:
      try:
        command.remove(flag)
      except ValueError:
        break
      count += 1
    if count != 0:
      flags.remove(flag)
  if flags:
    fmt = ' '.join(shell.quote(x) for x in flags)
    logger.warn("flags not removed: " + fmt)

  return command


@contextlib.contextmanager
def combine_context(*inputs):
  """
  Combines multiple context managers.
  """

  try:
    for ctx in inputs:
      ctx.__enter__()
    yield inputs
  finally:
    raise_later = []
    for ctx in inputs:
      try:
        ctx.__exit__(*sys.exc_info())
      except BaseException:
        raise_later.append(sys.exc_info())
    for exc_type, exc_value, exc_tb in raise_later:
      raise exc_value.with_traceback(exc_tb)


def copyfileobj(sfp, dfp, size=None, chunksize=4096):
  """
  Generator that yields the progress after every chunk has been copied
  from *sfp* to *dfp*. If *size* is not specified, *sfp* will be seeked
  first. The yielded progress is a tuple of ``(bytes_copied, size)``.
  """

  if size is None:
    pos = sfp.tell()
    sfp.seek(os.SEEK_END)
    size = sfp.tell()
    sfp.seek(pos)

  bytes_copied = 0
  while True:
    yield bytes_copied, size
    data = dfp.read(chunksize)
    if not data:
      break
    written = sfp.write(data)
    if written is not None and written != len(data):
      raise IOError('wrote {} of {} bytes'.format(written, len(data)))
    bytes_copied += len(data)



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr.utils import tty

import abc
import contextlib
import itertools
import sys
import time
import werkzeug

DEBUG = 5
INFO = 10
WARNING = 15
ERROR = 20

class BaseLogger(object, metaclass=abc.ABCMeta):

  DEBUG = DEBUG
  INFO = INFO
  WARNING = WARNING
  ERROR = ERROR

  def debug(self, *args, **kwargs):
    self.log(DEBUG, *args, **kwargs)

  def info(self, *args, **kwargs):
    self.log(INFO, *args, **kwargs)

  def warn(self, *args, **kwargs):
    self.log(WARNING, *args, **kwargs)

  def error(self, *args, **kwargs):
    self.log(ERROR, *args, **kwargs)

  @contextlib.contextmanager
  def indent(self):
    self.add_indent(1)
    try:
      yield
    finally:
      self.add_indent(-1)

  @abc.abstractmethod
  def log(self, level, *object, sep=' ', end='\n', indent=0):
    pass

  @abc.abstractmethod
  def add_indent(self, level):
    pass

  @abc.abstractmethod
  def progress_begin(self, description, spinning):
    pass

  @abc.abstractmethod
  def progress_update(self, progress, info_text=''):
    pass

  @abc.abstractmethod
  def progress_end(self):
    pass

  @abc.abstractmethod
  def set_level(self, level):
    pass


class DefaultLogger(BaseLogger):

  level_colors = {
    DEBUG: 'yellow',
    INFO: 'white',
    WARNING: 'magenta',
    ERROR: 'red'
  }

  def __init__(self, stream=None, level=INFO, indent_seq='  '):
    self._stream = stream or sys.stdout
    self._level = level
    self._indent_seq = indent_seq
    self._indent = 0
    self._progress = None
    self._line_alive = False

  def log(self, level, *objects, sep=' ', end='\n', indent=0):
    if level < self._level:
      return
    if self._progress:
      tty.clear_line()
    prefix = '' if self._line_alive else self._indent_seq * (self._indent + indent)
    prefix += tty.compile(self.level_colors[level])
    content = sep.join(map(str, objects))
    for line in content.split('\n'):
      print(prefix + line + tty.reset, end=end, file=self._stream)
    self._line_alive = ('\n' not in end)
    if self._progress and 'progress' in self._progress:
      self.progress_update(self._progress['progress'], self._progress['info_text'], _force=True)
    self._stream.flush()

  def add_indent(self, levels):
    self._indent += levels

  def progress_begin(self, description=None, spinning=False):
    self._progress = {'description': description, 'spinning': spinning,
      'index': 0, 'last': 0}
    if description:
      self.info(description)

  def progress_update(self, progress, info_text='', *, _force=False):
    if not self._progress:
      return
    info_text = str(info_text)
    self._progress['progress'] = progress
    self._progress['info_text'] = info_text
    ctime = time.time()
    if not _force and ctime - self._progress['last'] < 0.25:
      return
    tty.clear_line()
    width = 30
    if self._progress['spinning']:
      sign = ('~--', '-~-', '--~')[self._progress['index'] % 3]
      bar = ''.join(itertools.islice(itertools.cycle(sign), width))
    else:
      intprogress = int(min([1.0, max([0.0, float(progress)])]) * width)
      bar = '#' * intprogress + ' ' * (width - intprogress)
    indent = self._indent_seq * self._indent
    print('{}|{}| {}'.format(indent, bar, info_text), end='', file=self._stream)
    self._progress['index'] += 1
    self._progress['last'] = ctime
    self._stream.flush()

  def progress_end(self):
    tty.clear_line()
    self._progress = None

  def set_level(self, level):
    self._level = level


_logger = DefaultLogger()
logger = werkzeug.LocalProxy(lambda: _logger)


def set_logger(logger):
  global _logger
  _logger = logger



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr.utils import path

name = "mac"
standard = "posix"

def obj(x): return path.addsuffix(x, ".o")
def bin(x): return x
def dll(x): return path.addsuffix(x, ".dylib")
def lib(x): return path.addprefix(path.addsuffix(x, ".a"), "lib")



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['external_file', 'external_archive']

from craftr.defaults import buildlocal, gtn, logger, session, Framework, path, shell
from craftr.utils import httputils, pyutils

import nr.misc.archive


def get_loader_cache(loader_name, module=None):
  """
  Retrieve the cache of a loader by its *loader_name* and *module*. If no
  cache exists, an empty dictionary is returned.

  :param loader_name: The name of the loader.
  :param module: A :class:`~craftr.core.session.Module` object or a string.
  """

  if module is None:
    module = session.module
  if not isinstance(module, str):
    module = module.manifest.name

  loaders = session.cache.setdefault('loaders', {})
  module_cache = loaders.setdefault(module, {})
  return module_cache.setdefault(loader_name, {})


def _external_file_download_callback(progress_text, directory, filename, cache, data):
  """
  *Private*. Callback function used in :func:`external_file` when a file
  is downloaded with :func:`httputils.download_file`.

  :param progress_text: The text to display on the progress bar.
  :param directory: The directory passed to :func:`external_file`.
  :param filename: The filename passed to :func:`external_file`.
  :param cache: A dictionary that contains cached information, eventually
      from a previous download.
  :param data: The data passed from :func:`httputils.download_file`.
  """

  is_spinning = data['size'] is None

  # Before the download starts but the request is already made, we can
  # check if the file already exists.
  if data['downloaded'] == 0:
    logger.progress_begin(progress_text, is_spinning)
  elif data['completed']:
    logger.progress_end()
  elif is_spinning:
    # TODO: Convert bytes to human readable
    logger.progress_update(None, data['downloaded'])
  elif not is_spinning:
    # TODO: Convert bytes to human readable
    logger.progress_update(data['downloaded'] / data['size'], data['downloaded'])


class NoExternalFileMatch(Exception):

  def __init__(self, name, urls, excs):
    if len(urls) != len(excs):
      raise ValueError("number of exceptions must match number of urls")
    self.name = name
    self.urls = urls
    self.excs = excs

  def __str__(self):
    urls = ['  - {} ({})'.format(u, e) for (u, e) in zip(self.urls, self.excs)]
    return '\n'.join([self.name] + urls)


def external_file(*urls, filename = None, directory = None,
    copy_file_url = False, name = None):
  """
  Downloads a file from the first valid URL and saves it into *directory*
  under the specified *filename*.

  :param urls: One or more URLs. Supports ``http://``, ``https://``,
      ``ftp://`` and ``file://`. Note that if a ``file://`` URL is
      specified, the file is not copied to the output filename unless
      *copy_file_url* is True.
  :param filename: The output filename of the downloaded file. Defaults
      to the filename of the downloaded file.
  :param directory: The directory to save the file to. If *filename*
      is a relative path, it will be joined with this directory. Defaults
      to a path in the build directory.
  :param copy_file_url: If True, ``file://`` URLs will be copied instead
      of used as-is.
  :param name: The name of the loader action. This name is used to store
      information in the :attr:`Session.cache` so we can re-use existing
      downloaded data. :func:`~craftr.defaults.gtn` will be used to
      retrieve the default value for this parameter.
  :return: The path to the downloaded file.
  """

  name = gtn(name)
  if not directory and not filename:
    directory = buildlocal('data')

  cache = get_loader_cache(name)

  # TODO: expand variables of the current module.

  target_filename = None
  exceptions = []
  for url in urls:
    if url == cache.get('download_url'):
      existing_file = cache.get('download_file')
      if existing_file and path.isfile(existing_file):
        return existing_file

    progress_info = 'Downloading {} ...'.format(url)
    if url.startswith('file://'):
      source_file = url[7:]
      if path.isfile(source_file):
        if not copy_file_url:
          return source_file
        if not filename:
          filename = path.basename(source_file)

        # TODO: Use httputils.download_file() for this as well?
        logger.progress_begin(progress_info)
        path.makedirs(directory)
        target_filename = path.join(directory, filename)
        with open(source_file, 'rb') as sfp:
          with open(target_filename, 'wb') as dfp:
            for bytes_copied, size in pyutils.copyfileobj(sfp, dfp):
              logger.progress_update(float(bytes_copied) / size)
        logger.progress_end()

        # TODO: Copy file permissions
        break
      else:
        exceptions.append(FileNotFoundError(url))
    else:
      progress = lambda data: _external_file_download_callback(
          progress_info, directory, filename, cache, data)

      try:
        target_filename, reused = httputils.download_file(
          url, filename = filename, directory = directory, on_exists = 'skip',
          progress = progress)
      except (httputils.URLError, httputils.HTTPError) as exc:
        exceptions.append(exc)
      else:
        break
      finally:
        logger.progress_end()

  if target_filename:
    cache['download_url'] = url
    cache['download_file'] = target_filename
    return target_filename

  raise NoExternalFileMatch(name, urls, exceptions)


def external_archive(*urls, exclude_files = (), directory = None, name = None):
  """
  Downloads an archive from the first valid URL and unpacks it into
  *directory*. Archives with a single directory at the root will be
  extracted from one level below, eliminating that single parent
  directory.

  *exclude_files* can be a list of glob patterns that will be matched
  against the arcnames in the archive. Note that to exclude a directory,
  a pattern must match all files in that directory.

  Uses :func:`external_file` to download the archive.

  :param urls: See :func:`external_file`
  :param exclude_files: A list of glob patterns.
  :param directory: The directory to unpack the archive to. Defaults
      to a directory on the build directory derived from the downloaded
      archive filename. If defined and followed by a trailing slash, the
      archive filename will be appended.
  :param name: The name of the loader action. This name is used to store
      information in the :attr:`Session.cache` so we can re-use existing
      downloaded data. :func:`~craftr.defaults.gtn` will be used to
      retrieve the default value for this parameter.
  :return: The path to the top-level directory of the unpacked archive.
  """

  name = gtn(name)
  if not directory:
    directory = buildlocal('data') + '/'

  archive = external_file(*urls, directory = directory, name = name)
  cache = get_loader_cache(name)  # shared with external_file()

  suffix = nr.misc.archive.get_opener(archive)[0]
  if path.maybedir(directory):
    filename = path.basename(archive)[:-len(suffix)]
    directory = path.join(directory, filename)

  # Check if we already unpacked it etc.
  if cache.get('archive_source') == archive and \
      cache.get('archive_dir') == directory and \
      path.isdir(directory):
    return directory

  def match_exclude_files(arcname):
    for pattern in exclude_files:
      if fnmatch.fnmatch(arcname, pattern):
        return False
    return True

  def progress(index, count, filename):
    if index == -1:
      logger.progress_begin("Unpacking {} ...".format(path.basename(archive)))
      logger.progress_update(0.0, 'Reading index...')
      return
    progress = index / float(count)
    if index == 0:
      logger.progress_end()
      logger.progress_begin(None, False)
    elif index == (count - 1):
      logger.progress_end()
    else:
      logger.progress_update(progress, '{} / {}'.format(index, count))

  nr.misc.archive.extract(archive, directory, suffix = suffix,
    unpack_single_dir = True, check_extract_file = match_exclude_files,
    progress_callback = progress)

  cache['archive_source'] = archive
  cache['archive_dir'] = directory
  return directory


class PkgConfigError(Exception):
  pass


def pkg_config(pkg_name):
  """
  If available, this function uses ``pkg-config`` to extract flags for
  compiling and linking with the package specified with *pkg_name*. If
  ``pkg-config`` is not available or the it can not find the package,
  :class:`PkgConfigError` is raised.
  """

  try:
    flags = shell.pipe(['pkg-config', pkg_name, '--cflags', '--libs'], check = True).stdout
  except FileNotFoundError as exc:
    raise PkgConfigError('pkg-config is not available ({})'.format(exc))
  except shell.CalledProcessError as exc:
    raise PkgConfigError('{} not installed on this system\n\n{}'.format(
        pkg_name, exc.stderr or exc.stdout))

  # Parse the flags.
  result = Framework('pkg-config:{}'.format(pkg_name),
      include = [], defines = [], libs = [], libpath = [],
      compile_additional_flags = [], link_additional_flags = [])
  for flag in shell.split(flags):
    if flag.startswith('-I'):
      result['include'].append(flag[2:])
    elif flag.startswith('-D'):
      result['defines'].append(flag[2:])
    elif flag.startswith('-l'):
      result['libs'].append(flag[2:])
    elif flag.startswith('-L'):
      result['libpath'].append(flag[2:])
    elif flag.startswith('-Wl,'):
      result['link_additional_flags'].append(flag[4:])
    else:
      result['compile_additional_flags'].append(flag)

  return result


pkg_config.Error = PkgConfigError



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr import core
from craftr.core.config import read_config_file, InvalidConfigError
from craftr.core.logging import logger
from craftr.core.session import session, Session, Module, MANIFEST_FILENAME
from craftr.utils import path, shell
from nr.types.version import Version, VersionCriteria

import abc
import argparse
import atexit
import configparser
import craftr.defaults
import craftr.targetbuilder
import functools
import json
import os
import sys
import textwrap

CONFIG_FILENAME = '.craftrconfig'


def parse_cmdline_options(options):
  for item in options:
    key, sep, value = item.partition('=')
    if not sep:
      value = 'true'
    if not value:
      session.options.pop(key, None)
    else:
      session.options[key] = value


def read_cache(cachefile):
  try:
    with open(cachefile) as fp:
      try:
        session.read_cache(fp)
      except ValueError as exc:
        return False
  except IOError as exc:
    return False
  return True


def write_cache(cachefile):
  # Write back the cache.
  try:
    path.makedirs(path.dirname(cachefile))
    with open(cachefile, 'w') as fp:
      session.write_cache(fp)
  except OSError as exc:
    logger.error('error writing cache file:', cachefile)
    logger.error(exc, indent=1)
  else:
    logger.debug('cache written:', cachefile)


def parse_module_spec(spec):
  """
  Parses a module spec as it can be specified on the command-line. The
  format is ``module_name[:version]``.
  """

  parts = spec.split(':')
  if len(parts) not in (1, 2):
    raise ValueError('invalid module spec: {!r}'.format(spec))
  if len(parts) == 1:
    parts.append('*')
  try:
    version = Version(parts[1])
  except ValueError as exc:
    version = VersionCriteria(parts[1])
  return parts[0], version


def get_volatile_module_version(name):
  """
  Given a module name of which the exact version number may be appended
  and separated by a hyphen, returns the raw module name and its version
  as a tuple. If the module is not suffixed with a version number, the
  returned version number is :const:`None` instead.
  """

  parts = name.rpartition('-')[::2]
  if len(parts) != 2:
    return name
  try:
    version = Version(parts[1])
  except ValueError:
    version = None
  else:
    name = parts[0]
  return name, version


@functools.lru_cache()
def get_ninja_version(ninja_bin):
  ''' Read the ninja version from the `ninja` program and return it. '''

  return shell.pipe([ninja_bin, '--version'], shell=True).output.strip()


def get_ninja_info():
  # Make sure the Ninja executable exists and find its version.
  ninja_bin = session.options.get('global.ninja') or \
      session.options.get('craftr.ninja') or os.getenv('NINJA', 'ninja')
  ninja_bin = shell.find_program(ninja_bin)
  ninja_version = get_ninja_version(ninja_bin)
  logger.debug('Ninja executable:', ninja_bin)
  logger.debug('Ninja version:', ninja_version)
  return ninja_bin, ninja_version


class BaseCommand(object, metaclass=abc.ABCMeta):
  """
  Base class for Craftr subcommands.
  """

  @abc.abstractmethod
  def build_parser(self, parser):
    pass

  @abc.abstractmethod
  def execute(self, parser, args):
    pass



class ExportOrBuildCommand(BaseCommand):

  def __init__(self, is_export):
    self.is_export = is_export

  def build_parser(self, parser):
    if self.is_export:
      parser.add_argument('-m', '--module')
    else:
      parser.add_argument('targets', metavar='TARGET', nargs='*')
    parser.add_argument('-d', '--option', dest='options', action='append', default=[])
    parser.add_argument('-b', '--build-dir', default='build')
    parser.add_argument('-i', '--include-path', action='append', default=[])

  def execute(self, parser, args):
    session.path.extend(map(path.norm, args.include_path))

    if self.is_export:
      # Determine the module to execute, either from the current working
      # directory or find it by name if one is specified.
      if not args.module:
        for fn in [MANIFEST_FILENAME, path.join('craftr', MANIFEST_FILENAME)]:
          if path.isfile(fn):
            module = session.parse_manifest(fn)
            break
        else:
          parser.error('"{}" does not exist'.format(MANIFEST_FILENAME))
      else:
        # TODO: For some reason, prints to stdout are not visible here.
        # TODO: Prints to stderr however work fine.
        try:
          module_name, version = parse_module_spec(args.module)
        except ValueError as exc:
          parser.error('{} (note: you have to escape > and < characters)'.format(exc))
        try:
          module = session.find_module(module_name, version)
        except Module.NotFound as exc:
          parser.error('module not found: ' + str(exc))
    else:
      module = None

    ninja_bin, ninja_version = get_ninja_info()

    # Create and switch to the build directory.
    session.builddir = path.abs(args.build_dir)
    path.makedirs(session.builddir)
    os.chdir(session.builddir)

    # Read the cache and parse command-line options.
    cachefile = path.join(session.builddir, '.craftrcache')
    if not read_cache(cachefile) and not self.is_export:
      logger.error('Unable to load "{}", can not build'.format(cachefile))
      return 1

    # Prepare options, loaders and execute.
    if self.is_export:
      session.expand_relative_options(module.manifest.name)
      session.cache['build'] = {}
      try:
        write_cache(cachefile)
        module.run()
      except Module.InvalidOption as exc:
        for error in exc.format_errors():
          logger.error(error)
        return 1
      except craftr.defaults.ModuleError as exc:
        logger.error('error:', exc)
        return 1

      # Write the cache back.
      session.cache['build']['targets'] = list(session.graph.targets.keys())
      session.cache['build']['main'] = module.ident
      session.cache['build']['options'] = args.options
      write_cache(cachefile)

      # Write the Ninja manifest.
      with open("build.ninja", 'w') as fp:
        platform = core.build.get_platform_helper()
        context = core.build.ExportContext(ninja_version)
        writer = core.build.NinjaWriter(fp)
        session.graph.export(writer, context, platform)

    else:
      parse_cmdline_options(session.cache['build']['options'])
      main = session.cache['build']['main']
      available_targets = frozenset(session.cache['build']['targets'])

      logger.debug('build main module:', main)
      session.expand_relative_options(get_volatile_module_version(main)[0])

      # Check the targets and if they exist.
      targets = []
      for target in args.targets:
        if '.' not in target:
          target = main + '.' + target
        elif target.startswith('.'):
          target = main + target

        module_name, target = target.rpartition('.')[::2]
        module_name, version = get_volatile_module_version(module_name)
        ref_module = session.find_module(module_name, version or '*')
        target = craftr.targetbuilder.get_full_name(target, ref_module)
        if target not in available_targets:
          parser.error('no such target: {}'.format(target))
        targets.append(target)

      # Execute the ninja build.
      cmd = [ninja_bin]
      if args.verbose:
        cmd += ['-v']
      cmd += targets
      shell.run(cmd)


class StartpackageCommand(BaseCommand):

  def build_parser(self, parser):
    parser.add_argument('name')
    parser.add_argument('directory', nargs='?', default=None)
    parser.add_argument('-n', '--nested', action='store_true')
    parser.add_argument('--version', type=Version, default='1.0.0')

  def execute(self, parser, args):
    directory = args.directory or args.name
    if path.maybedir(directory):
      directory = path.join(directory, args.name)

    if not path.exists(directory):
      logger.debug('creating directory "{}"'.format(directory))
      path.makedirs(directory)
    elif not path.isdir(directory):
      logger.error('"{}" is not a directory'.format(directory))
      return 1

    if args.nested:
      directory = path.join(directory, 'craftr')
      path.makedirs(directory)

    mfile = path.join(directory, MANIFEST_FILENAME)
    sfile = path.join(directory, 'Craftrfile')
    for fn in [mfile, sfile]:
      if path.isfile(fn):
        logger.error('"{}" already exists'.format(fn))
        return 1


    logger.debug('creating file "{}"'.format(mfile))
    with open(mfile, 'w') as fp:
      lines = textwrap.dedent('''
        {
          "name": "%s",
          "version": "%s",
          "project_dir": "..",
          "author": "",
          "url": "",
          "dependencies": {},
          "options": {}
        }\n''' % (args.name, args.version)).lstrip().split('\n')
      if not args.nested:
        del lines[3]
      fp.write('\n'.join(lines))

    logger.debug('creating file "{}"'.format(sfile))
    with open(sfile, 'w') as fp:
      print('# {}'.format(args.name), file=fp)


class version(BaseCommand):

  def build_parser(self, parser):
    pass

  def execute(self, parser, args):
    print(craftr.__version__)


def main():
  # Create argument parsers and dynamically include all BaseCommand
  # subclasses into it.
  parser = argparse.ArgumentParser(prog='craftr', description='The Craftr build system')
  parser.add_argument('-v', '--verbose', action='store_true')
  parser.add_argument('-q', '--quiet', action='store_true')
  parser.add_argument('-c', '--config', action='append', default=[])
  parser.add_argument('-C', '--no-config', action='store_true')
  parser.add_argument('-d', '--option', dest='options', action='append', default=[])
  subparsers = parser.add_subparsers(dest='command')

  commands = {
    'export': ExportOrBuildCommand(is_export=True),
    'build': ExportOrBuildCommand(is_export=False),
    'startpackage': StartpackageCommand()
  }
  for key, cmd in commands.items():
    cmd.build_parser(subparsers.add_parser(key))

  # Parse the arguments.
  args = parser.parse_args()
  if not args.command:
    parser.print_usage()
    return 0

  if args.verbose:
    logger.set_level(logger.DEBUG)
  elif args.quiet:
    logger.set_level(logger.WARNING)

  session = Session()

  # Parse the user configuration file.
  try:
    config_filename = path.expanduser('~/' + CONFIG_FILENAME)
    session.options = read_config_file(config_filename)
  except FileNotFoundError as exc:
    session.options = {}
  except InvalidConfigError as exc:
    parser.error(exc)
    return 1

  # Parse the local configuration file or the ones specified on command-line.
  if not args.no_config:
    try:
      for filename in args.config:
        session.options.update(read_config_file(filename))
      if not args.config:
        choices = [CONFIG_FILENAME, path.join('craftr', CONFIG_FILENAME)]
        for fn in choices:
          try:
            session.options.update(read_config_file(fn))
          except FileNotFoundError as exc:
            pass
    except InvalidConfigError as exc:
      parser.error(exc)
      return 1

  # Execute the command in the session context.
  with session:
    parse_cmdline_options(args.options)
    return commands[args.command].execute(parser, args)


def main_and_exit():
  sys.exit(main())


if __name__ == '__main__':
  main_and_exit()



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr.utils import argspec
from os import sep, pathsep, curdir, pardir, getcwd
from os.path import exists, isdir, isfile, isabs, abspath as abs
from os.path import join, split, dirname, basename, expanduser

import ctypes
import errno
import glob2
import os
import shutil
import tempfile as _tempfile

curdir_sep = curdir + sep
pardir_sep = pardir + sep


def rel(path, parent=None, nopar=False):
  """
  Like :func:`os.path.relpath`, but the *nopar* parameter can be set to return
  an absolute path if the relative path would create a path element that
  references a parent directory (`..`) or current directory (``.``).
  """

  try:
    res = os.path.relpath(path, parent)
  except ValueError:
    if nopar:
      return abs(path)
    raise
  else:
    if not issub(res):
      return abs(path)
    return res

def norm(path, parent=None):
  """
  Normalizes the specified *path*. This turns it into an absolute path and
  removes all superfluous path elements. Similar to :func:`os.path.normpath`,
  but accepts a *parent* argument which is considered when *path* is relative.
  """

  if not isabs(path):
    path = join(parent or getcwd(), path)
  return canonical(path)

def canonical(path):
  """
  A synonym for :meth:`os.path.normpath`.
  """

  return os.path.normpath(path)

def glob(patterns, parent=None, excludes=(), include_dotfiles=False):
  """
  Wrapper for :func:`glob2.glob` that accepts an arbitrary number of
  patterns and matches them. The paths are normalized with :func:`norm`.

  Relative patterns are automaticlly joined with *parent*. If the
  parameter is omitted, it defaults to the currently executed build
  scripts project directory.

  If *excludes* is specified, it must be a string or a list of strings
  that is/contains glob patterns or filenames to be removed from the
  result before returning.

  .. note::

    Every file listed in *excludes* will only remove **one** item from
    the result list that was generated from *patterns*. Thus, if you
    want to exclude some files with a pattern except for a specific file
    that would also match that pattern, simply list that file another
    time in the *patterns*.

  :param patterns: A list of glob patterns or filenames.
  :param parent: The parent directory for relative paths.
  :param excludes: A list of glob patterns or filenames.
  :param include_dotfiles: If True, ``*`` and ``**`` can also capture
    file or directory names starting with a dot.
  :return: A list of filenames.
  """

  argspec.validate('patterns', patterns, {'type': [list, tuple]})
  argspec.validate('excludes', excludes, {'type': [list, tuple]})
  argspec.validate('parent', parent, {'type': [None, str]})

  if not parent:
    parent = getcwd()

  result = []
  for pattern in patterns:
    if not isabs(pattern):
      pattern = join(parent, pattern)
    result += glob2.glob(norm(pattern))

  for pattern in excludes:
    if not isabs(pattern):
      pattern = join(parent, pattern)
    pattern = norm(pattern)
    if not isglob(pattern):
      result.remove(pattern)
    else:
      for item in glob2.glob(pattern):
        result.remove(item)

  return result

def isglob(path):
  """
  :param path: The string to check
  :return bool: True if the path is a glob pattern, False otherwise.
  """

  return '*' in path or '?' in path

def issub(path):
  """
  Returns True if *path* is a relative path that does not point outside
  of its parent directory or is equal to its parent directory (thus, this
  function will also return False for a path like ``./``).
  """

  if isabs(path):
    return False
  if path.startswith(curdir_sep) or path.startswith(pardir_sep):
    return False
  return True

def maybedir(path):
  """
  Returns True if *path* ends with a separator. This information can be used
  to interpret a path as a directory or a filename only from its
  representation.

  .. code:: python

    >>> maybedir('foobar/')
    True
    >>> maybedir('foo/bar')
    False
  """

  if os.name == 'nt':
    return path.endswith('/') or path.endswith('\\')
  return path.endswith('/')

def addprefix(subject, prefix):
  """
  Adds the specified *prefix* to the last path element in *subject*.
  If *prefix* is a callable, it must accept exactly one argument, which
  is the last path element, and return a modified value.
  """

  if not prefix:
    return subject
  dir_, base = split(subject)
  if callable(prefix):
    base = prefix(base)
  else:
    base = prefix + base
  return join(dir_, base)

def addsuffix(subject, suffix, replace=False):
  """
  Adds the specified *suffix* to the *subject*. If *replace* is True, the
  old suffix will be removed first. If *suffix* is callable, it must accept
  exactly one argument and return a modified value.
  """

  if not suffix and not replace:
    return subject
  if replace:
    subject = rmvsuffix(subject)
  if suffix and callable(suffix):
    subject = suffix(subject)
  elif suffix:
    subject += suffix
  return subject

def setsuffix(subject, suffix):
  """
  Synonymous for passing the True for the *replace* parameter in
  :func:`addsuffix`.
  """
  return addsuffix(subject, suffix, replace=True)

def rmvsuffix(subject):
  """
  Remove the suffix from *subject*.
  """

  index = subject.rfind('.')
  if index > subject.replace('\\', '/').rfind('/'):
    subject = subject[:index]
  return subject

def makedirs(path):
  """
  Like :func:`os.makedirs`, but this function does not raise an exception
  when the directory at *path* already exists.
  """

  try:
    os.makedirs(path)
  except FileExistsError:
    pass  # intentional

def remove(path, recursive=False, silent=False):
  """
  Like :func:`os.remove`, but the *silent* parmaeter can be specified to
  prevent the function from raising an exception if the *path* could not be
  removed. Also, the *recursive* parameter allows you to use this function
  to remove directories as well. In this case, :func:`shutil.rmtree` is
  used.
  """

  try:
    if recursive and isdir(path):
      shutil.rmtree(path)
    else:
      os.remove(path)
  except OSError as exc:
    if not silent or exc.errno != errno.ENOENT:
      raise

def get_long_path_name(path):
  """
  This function is important when using Craftr on platforms with
  case-insenstive filesystem, such as Windows. It returns the correct
  capitalization for *path*, given that it exists. For other platforms,
  the *path* is returned as-is.

  If *path* does not exist, it is returned unchanged in any case.
  """

  # TODO: Is Cygwin case-insenstive as well?

  if os.name == 'nt':
    # Thanks to http://stackoverflow.com/a/3694799/791713
    buf = ctypes.create_unicode_buffer(len(path) + 1)
    GetLongPathNameW = ctypes.windll.kernel32.GetLongPathNameW
    res = GetLongPathNameW(path, buf, len(path) + 1)
    if res == 0 or res > 260:
      return path
    else:
      return buf.value

  return path

def transition(filename, oldbase, newbase):
  """
  Translates the *filename* from the directory *oldbase* to the new
  directory *newbase*. This is identical to finding the relative path
  of *filename* to *oldbase* and joining it with *newbase*. The #filename
  must be a sub-path of *oldbase*.

  .. code:: python

    >>> transition('src/main.c', 'src', 'build/obj')
    build/obj/main.c
  """

  rel_file = rel(filename, oldbase, nopar=True)
  if isabs(rel_file):
    raise ValueError("filename must be a sub-path of oldbase", filename, oldbase)
  return join(newbase, rel_file)

def common(paths):
  """
  Returns the longest sub-path of each path in the *paths* list. If *paths* is
  empty, contains mixed absolute/relative paths or the paths have no common
  path, a :class:`ValueError` is raised.

  If there is only one element in *paths*, its parent directory is returned.
  """

  if not paths:
    raise ValueError('paths is empty')

  parts = []
  has_abs = None
  for path in paths:
    if not isabs(path):
      if has_abs is None:
        has_abs = False
      elif has_abs:
        raise ValueError('paths contains relative and absolute pathnames')
    else:
      if has_abs is None:
        has_abs = True
      elif not has_abs:
        raise ValueError('paths contains relative and absolute pathnames')

    path = norm(path)
    parts.append(path.split(sep))

  if len(parts) == 1:
    path = dirname(sep.join(parts[0]))
    if not has_abs:
      path = rel(path)
    return path

  common = parts[0]
  for elements in parts[1:]:
    if len(elements) < len(common):
      common = common[:len(elements)]
    for index, elem in enumerate(elements):
      if index >= len(common):
        break
      if elem != common[index]:
        common = common[:index]
        break
      if not common:
        break

  if not common:
    raise ValueError("no common path")

  common = sep.join(common)
  if not has_abs:
    common = rel(common)
  return common

def easy_listdir(directory):
  """
  A friendly version of :func:`os.listdir` that does not error if the
  *directory* doesn't exist.
  """

  try:
    return os.listdir(directory)
  except OSError as exc:
    if exc.errno != errno.ENOENT:
      raise
  return []

class tempfile(object):
  """
  A better temporary file class where the #close() function does not delete
  the file but only #__exit__() does. Obviously, this allows you to close
  the file and re-use it with some other processing before it finally gets
  deleted.

  This is especially important on Windows because apparently another
  process can't read the file while it's still opened in the process
  that created it.

  ```python
  from craftr.tools import tempfile
  with tempfile(suffix='c', text=True) as fp:
    fp.write('#include <stdio.h>\nint main() { }\n')
    fp.close()
    shell.run(['gcc', fp.name])
  ```

  @param suffix: The suffix of the temporary file.
  @param prefix: The prefix of the temporary file.
  @param dir: Override the temporary directory.
  @param text: True to open the file in text mode. Otherwise, it
    will be opened in binary mode.
  """

  def __init__(self, suffix='', prefix='tmp', dir=None, text=False):
    self.fd, self.name = _tempfile.mkstemp(suffix, prefix, dir, text)
    self.fp = os.fdopen(self.fd, 'w' if text else 'wb')

  def __enter__(self):
    return self

  def __exit__(self, *__):
    try:
      self.close()
    finally:
      remove(self.name, silent=True)

  def __getattr__(self, name):
    return getattr(self.fp, name)

  def close(self):
    if self.fp:
      self.fp.close()
      self.fp = None



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pip.req import parse_requirements
from setuptools import setup, find_packages

import functools
import os
import pip
import sys

if sys.version < '3.4' or sys.version >= '3.6':
  print('-----------------------------------------------------------------')
  print("WARNING: Craftr officially supports Python 3.4, 3.5")
  print("WARNING: Your current version is Python {}".format(sys.version[:5]))
  print('-----------------------------------------------------------------')

# parse_requirements() interface has changed in Pip 6.0
if pip.__version__ >= '6.0':
  parse_requirements = functools.partial(
      parse_requirements, session=pip.download.PipSession())

# Convert README.md to reST.
if os.path.isfile('README.md'):
  if os.system('pandoc -s README.md -o README.rst') != 0:
    print('-----------------------------------------------------------------')
    print('WARNING: README.rst could not be generated, pandoc command failed')
    print('-----------------------------------------------------------------')
    if sys.stdout.isatty():
      input("Enter to continue... ")

if os.path.isfile('README.rst'):
  with open('README.rst', encoding='utf8') as fp:
    long_description = fp.read()
else:
  long_description = None

setup(
  name = 'craftr-build',
  version = '2.0.0.dev3',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Meta build system based on Ninja and Python',
  long_description = long_description,
  url = 'https://gitlab.niklasrosenstein.com/niklas/craftr',
  install_requires = [str(x.req) for x in parse_requirements('requirements.txt')],
  entry_points = dict(
    console_scripts = [
      'craftr = craftr.__main__:main_and_exit'
    ]
  ),
  packages = find_packages(),
  package_data = {
    'craftr': ['stl/**/*', 'stl_auxiliary/**/*']
  },
  license = 'MIT',
  classifiers = [
    "Topic :: Software Development",
    "Topic :: Software Development :: Build Tools",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: Implementation :: CPython",
    "License :: OSI Approved :: MIT License"
  ],
)



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from craftr.core.logging import logger
from craftr.utils import path

import configparser
import re


def read_config_file(filename, basedir=None, follow_include_directives=True):
  """
  Reads a configuration file and returns a dictionary of the values that
  it contains. The format is standard :mod:`configparser` ``.ini`` style,
  however this function supports ``include`` directives that can include
  additional configuration files.

  ::

    [include "path/to/config.ini"]            ; errors if the file does not exist
    [include "path/to/config.ini" if-exists]  ; ignored if the file does not exist

  :param filename: The name of the configuration file to read.
  :param basedir: If *filename* is not an absolute path or the base directory
    should be altered, this is the directory of which to look for files
    specified with ``include`` directives.
  :param follow_include_directives: If this is True, ``include`` directives
    will be followed.
  :raise FileNotFoundError: If *filename* does not exist.
  :raise InvalidConfigError: If the configuration format is invalid. Also
    if any of the included files do not exist.
  :return: A dictionary. Section names are prepended to the option names.
  """

  filename = path.norm(filename)
  if not basedir:
    basedir = path.dirname(filename)

  logger.debug('reading configuration file:', filename)
  if not path.isfile(filename):
    raise FileNotFoundError(filename)
  parser = configparser.SafeConfigParser()
  try:
    parser.read([filename])
  except configparser.Error as exc:
    raise InvalidConfigError('"{}": {}'.format(filename, exc))

  result = {}
  for section in parser.sections():
    match = re.match('include\s+"([^"]+)"(\s+if-exists)?$', section)
    if match:
      if not follow_include_directives:
        continue
      ifile, if_exists = match.groups()
      ifile = path.norm(ifile, basedir)
      try:
        result.update(read_config_file(ifile))
      except FileNotFoundError as exc:
        if not if_exists:
          raise InvalidConfigError('file "{}" included by "{}" does not exist'
              .format(str(exc), filename))
      continue
    elif section == '__global__':
      prefix = ''
    else:
      prefix = section + '.'

    for option in parser.options(section):
      result[prefix + option] = parser.get(section, option)

  return result


class InvalidConfigError(Exception):
  pass



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

options = session.module.options
from craftr.loaders import external_archive

source_directory = external_archive(
  "https://github.com/jpbarrette/curlpp/archive/{}.zip".format(options.version)
)

if options.static:
  session.options.setdefault('craftr.lib.curl.static', True)
else:
  session.options.setdefault('craftr.lib.curl.static', False)

load_module('craftr.lang.cxx.*')

cURL = load_module('craftr.lib.curl').cURL
cURLpp = Framework('cURLpp',
  include = [path.join(source_directory, 'include')],
  defines = [],
  frameworks = [cURL]
)

if options.static:
  cURLpp['defines'] += ['CURLPP_STATICLIB']

cURLpp_library = cxx_library(
  link_style = 'static' if options.static else 'shared',
  inputs = cpp_compile(
    sources = glob(['src/**/*.cpp'], parent = source_directory),
    frameworks = [cURLpp],
    defines = ['BUILDING_CURLPP'],
    rtti = options.rtti
  ),
  output = 'cURLpp'
)

cxx_extend_framework(cURLpp, cURLpp_library)




# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections

def tn(value):
  return type(value).__name__

def validate(name, value, schema):
  """
  A helper function to validate function parameters type and value.
  The *schema* must be a dictionary and supports the following keys:

  - ``type``: A single type or a list of accepted types
  - ``bool_validators``: A function or a list of functions that return
    True if the *value* is valid, False otherwise.
  - ``validators``: A function or a list of functions that raise a
    :class:`ValueError` if the *value* is invalid.
  - ``items``: If *value* is a sequence, this key must provide a sub-schema
    that holds true for all of the items in the sequence. Note that it will
    not be applied to iterables.
  - ``allowEmpty``: If specified, must be True or False. If True, allows
    *value* to be an empty sequence, otherwise not.
  """

  schema.setdefault('type', [])
  schema.setdefault('bool_validators', [])
  schema.setdefault('validators', [])

  if isinstance(schema['type'], list):
    schema['type'] = tuple(schema['type'])
  elif not isinstance(schema['type'], tuple):
    schema['type'] = (schema['type'],)
  schema['type'] = tuple(type(None) if x is None else x for x in schema['type'])

  if schema['type'] and not isinstance(value, schema['type']):
    raise TypeError("argument '{}' expected one of {} but got {}".format(
      name, '{'+','.join(x.__name__ for x in schema['type'])+'}', tn(value)))
  if isinstance(value, collections.Sequence):
    if 'items' in schema:
      for index, item in enumerate(value):
        validate('{}[{}]'.format(name, index), item, schema['items'])
    if not schema.get('allowEmpty', True) and not value:
      raise ValueError("argument '{}' can not be empty".format(name))

  if not isinstance(schema['bool_validators'], (list, tuple)):
    schema['bool_validators'] = [schema['bool_validators']]
  for validator in schema['bool_validators']:
    if not validator(value):
      raise TypeError("argument '{}' is not {}".format(name, validator.__name__))

  if not isinstance(schema['validators'], (list, tuple)):
    schema['validators'] = [schema['validators']]
  for validator in schema['validators']:
    validator(value)



# The Craftr build system
# Copyright (C) 2016  Niklas Rosenstein
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`craftr.core.session`
==========================

This module provides the :class:`Session` class which manages the loading
process of Craftr modules and contains all the important root datastructures
for the meta build process (such as a :class:`craftr.core.build.Graph`).
"""

from craftr.core import build, manifest
from craftr.core.logging import logger
from craftr.core.manifest import Manifest
from craftr.utils import argspec, path
from nr.types.version import Version, VersionCriteria

import json
import os
import tempfile
import types
import werkzeug

MANIFEST_FILENAME = 'manifest.json'


class ModuleNotFound(Exception):

  def __init__(self, name, version):
    self.name = name
    self.version = version

  def __str__(self):
    if isinstance(self.version, Version):
      return '{}[={}]'.format(self.name, self.version)
    else:
      return '{}[{}]'.format(self.name, self.version)


class InvalidOption(Exception):

  def __init__(self, module, errors):
    self.module = module
    self.errors = errors

  def __str__(self):
    return '\n'.join(self.format_errors())

  def format_errors(self):
    for option, value, exc in self.errors:
      yield '{}.{} ({}): {}'.format(self.module.manifest.name, option.name,
          self.module.manifest.version, exc)


class Session(object):
  """
  This class manages the :class:`build.Graph` and loading of Craftr modules.

  .. attribute:: graph

    A :class:`build.Graph` instance.

  .. attribute:: path

    A list of paths that will be searched for Craftr modules.

  .. attribute:: require

    An instance of :class:`Require` that is used to load other Craftr
    modules conveniently.

  .. attribute:: module

    The Craftr module that is currently being executed. This is an instance
    of the :class:`Module` class and the same as the tip of the
    :attr:`modulestack`.

  .. attribute:: modulestack

    A list of modules where the last element (tip) is the module that is
    currently being executed.

  .. attribute:: modules

    A nested dictionary that maps from name to a dictionary of version
    numbers mapping to :class:`Module` objects. These are the modules that
    have already been loaded into the session or that have been found and
    cached but not yet been executed.

  .. attribute:: maindir

    The main directory from which Craftr was run. Craftr will switch to the
    build directory at a later point, which is why we keep this member for
    reference.

  .. attribute:: builddir

    The absolute path to the build directory.

  .. attribute:: options

    A dictionary of options that are passed down to Craftr modules.

  .. attributes:: cache

    A JSON object that will be loaded from the current workspace's cache
    file and written back when Craftr exits without errors. The cache can
    contain anything and can be modified by everything, however it should
    be assured that no name conflicts and accidental modifications/deletes
    occur.

    Reserved keywords in the cache are ``"build"`` and ``"loaders"``.
  """

  #: The current session object. Create it with :meth:`start` and destroy
  #: it with :meth:`end`.
  current = None

  #: Diretory that contains the Craftr standard library.
  stl_dir = path.norm(path.join(__file__, '../../stl'))
  stl_auxiliary_dir = path.norm(path.join(__file__, '../../stl_auxiliary'))

  def __init__(self, maindir=None):
    self.maindir = path.norm(maindir or path.getcwd())
    self.builddir = path.join(self.maindir, 'build')
    self.graph = build.Graph()
    self.path = [self.stl_dir, self.stl_auxiliary_dir, self.maindir,
        path.join(self.maindir, 'craftr/modules')]
    self.modulestack = []
    self.modules = {}
    self.options = {}
    self.cache = {}
    self._tempdir = None
    self._manifest_cache = {}  # maps manifest_filename: manifest
    self._refresh_cache = True

  def __enter__(self):
    if Session.current:
      raise RuntimeError('a session was already created')
    Session.current = self
    return Session.current

  def __exit__(self, exc_value, exc_type, exc_tb):
    if Session.current is not self:
      raise RuntimeError('session not in context')
    if self._tempdir and not self.options.get('craftr.keep_temporary_directory'):
      logger.debug('removing temporary directory:', self._tempdir)
      try:
        path.remove(self._tempdir, recursive=True)
      except OSError as exc:
        logger.debug('error:', exc, indent=1)
      finally:
        self._tempdir = None
    Session.current = None

  @property
  def module(self):
    if self.modulestack:
      return self.modulestack[-1]
    return None

  def read_cache(self, fp):
    cache = json.load(fp)
    if not isinstance(cache, dict):
      raise ValueError('Craftr Session cache must be a JSON object, got {}'
          .format(type(cache).__name__))
    self.cache = cache

  def write_cache(self, fp):
    json.dump(self.cache, fp, indent='\t')

  def expand_relative_options(self, main_module):
    """
    After the main module has been detected, relative option names (starting
    with ``.``) should be converted to absolute option names. This is what
    the method does.

    :param main_module: The name of the main module.
    """

    for key in tuple(self.options.keys()):
      if key.startswith('.'):
        self.options[main_module + key] = self.options.pop(key)

  def get_temporary_directory(self):
    """
    Returns a writable temporary directory that is primarily used by loaders
    to store temporary files. The temporary directory will be deleted when
    the Session context ends unless the ``craftr.keep_temporary_directory``
    option is set.

    :raise RuntimeError: If the session is not currently in context.
    """

    if Session.current is not self:
      raise RuntimeError('session not in context')
    if not self._tempdir:
      self._tempdir = path.join(self.builddir, '.temp')
      logger.debug('created temporary directory:', self._tempdir)
    return self._tempdir

  def parse_manifest(self, filename):
    """
    Parse a manifest by filename and add register the module to the module
    cache. Returns the :class:`Module` object. If the manifest has already
    been parsed, it will not be re-parsed.

    :raise Manifest.Invalid: If the manifest is invalid.
    :return: :const:`None` if the manifest is a duplicate of an already
      parsed manifest (determined by name and version), otherwise the
      :class:`Module` object for the manifest's module.
    """

    filename = path.norm(path.abs(filename))
    if filename in self._manifest_cache:
      manifest = self._manifest_cache[filename]
      return self.find_module(manifest.name, manifest.version)

    manifest = Manifest.parse(filename)
    self._manifest_cache[filename] = manifest
    versions = self.modules.setdefault(manifest.name, {})
    if manifest.version in versions:
      logger.debug('multiple occurences of "{}-{}" found, '
          'one of which is located at "{}"'.format(manifest.name,
          manifest.version, filename))
      module = None
    else:
      logger.debug('parsed manifest: {}-{} ({})'.format(
          manifest.name, manifest.version, filename))
      module = Module(path.dirname(filename), manifest)
      versions[manifest.version] = module

    return module

  def update_manifest_cache(self, force=False):
    if not self._refresh_cache and not force:
      return
    self._refresh_cache = False

    for directory in self.path:
      choices = []
      choices.append(path.join(directory, MANIFEST_FILENAME))
      for item in path.easy_listdir(directory):
        choices.append(path.join(directory, item, MANIFEST_FILENAME))
        choices.append(path.join(directory, item, 'craftr', MANIFEST_FILENAME))

      for filename in map(path.norm, choices):
        if filename in self._manifest_cache:
          continue  # don't parse a manifest that we already parsed
        if not path.isfile(filename):
          continue
        try:
          self.parse_manifest(filename)
        except Manifest.Invalid as exc:
          logger.warn('invalid manifest found:', filename)
          logger.warn(exc, indent=1)

  def find_module(self, name, version):
    """
    Finds a module in the :attr:`path` matching the specified *name* and
    *version*.

    :param name: The name of the module.
    :param version: A :class:`VersionCriteria`, :class:`Version` or string
      in a VersionCritiera format.
    :raise ModuleNotFound: If the module can not be found.
    :return: :class:`Module`
    """

    argspec.validate('name', name, {'type': str})
    argspec.validate('version', version, {'type': [str, Version, VersionCriteria]})

    if isinstance(version, str):
      try:
        version = Version(version)
      except ValueError as exc:
        version = VersionCriteria(version)

    self.update_manifest_cache()
    if name in self.modules:
      if isinstance(version, Version):
        if version in self.modules[name]:
          return self.modules[name][version]
        raise ModuleNotFound(name, version)
      for module in sorted(self.modules[name].values(),
          key=lambda x: x.manifest.version, reverse=True):
        if version(module.manifest.version):
          return module

    raise ModuleNotFound(name, version)


class Module(object):
  """
  This class represents a Craftr module that has been or is currently being
  executed. Every module has a project directory and a manifest with some
  basic information on the module such as its name, version, but also things
  like its dependencies and options.

  Every Craftr project (i.e. module) contains a ``manifest.json`` file
  and the main ``Craftrfile``.

  ::

    myproject/
      include/
      source/
      Craftrfile
      manifest.json

  .. attribute:: directory

    The directory that contains the ``manifest.json``. Note that the actual
    project directory depends on the :attr:`Manifest.project_dir` member.

  .. attribute:: ident

    A concentation of the name and version defined in the :attr:`manifest`.

  .. attribute:: project_dir

    Path to the project directory as specified in the :attr:`manifest`.

  .. attribute:: manifest

  .. attribute:: namespace

  .. attribute:: executed

    True if the module was executed with :meth:`run`.

  .. attribute:: options

    A :class:`~craftr.core.manifest.Namespace` that contains all the options
    for the module. This member is only initialized when the module is run
    or with :meth:`init_options`.
  """

  NotFound = ModuleNotFound
  InvalidOption = InvalidOption

  def __init__(self, directory, manifest):
    self.directory = directory
    self.manifest = manifest
    self.namespace = types.ModuleType(self.manifest.name)
    self.executed = False
    self.options = None

  def __repr__(self):
    return '<craftr.core.session.Module "{}-{}">'.format(self.manifest.name,
      self.manifest.version)

  @property
  def ident(self):
    return '{}-{}'.format(self.manifest.name, self.manifest.version)

  @property
  def project_dir(self):
    return path.norm(path.join(self.directory, self.manifest.project_dir))

  def init_options(self, recursive=False, _break_recursion=None):
    """
    Initialize the :attr:`options` member. Requires an active session context.

    :param recursive: Initialize the options of all dependencies as well.
    :raise InvalidOption: If one or more options are invalid.
    :raise ModuleNotFound: If *recursive* is specified and a dependency
      was not found.
    :raise RuntimeError: If there is no current session context.
    """

    if not session:
      raise RuntimeError('no current session')
    if _break_recursion is self:
      return

    if recursive:
      for name, version in self.manifest.dependencies.items():
        module = session.find_module(name, version)
        module.init_options(True, _break_recursion=self)

    if self.options is None:
      errors = []
      self.options = self.manifest.get_options_namespace(session.options, errors)
      if errors:
        self.options = None
        raise InvalidOption(self, errors)

  def run(self):
    """
    Loads the code of the main Craftr build script as specified in the modules
    manifest and executes it. Note that this must occur in a context where
    the :data:`session` is available.

    :raise RuntimeError: If there is no current :data:`session` or if the
      module was already executed.
    """

    if not session:
      raise RuntimeError('no current session')
    if self.executed:
      raise RuntimeError('already run')

    self.executed = True
    self.init_options()

    script_fn = path.norm(path.join(self.directory, self.manifest.main))
    with open(script_fn) as fp:
      code = compile(fp.read(), script_fn, 'exec')

    vars(self.namespace).update(self.get_init_globals())
    self.namespace.__file__ = script_fn
    self.namespace.__name__ = self.manifest.name
    self.namespace.__version__ = str(self.manifest.version)
    try:
      session.modulestack.append(self)
      exec(code, vars(self.namespace))
    finally:
      assert session.modulestack.pop() is self

  def get_init_globals(self):
    """
    Returns a dictionary initialized with the default built-in values for a
    build script.
    """

    from craftr import defaults
    result = {}
    for key, value in vars(defaults).items():
      if not key.startswith('_'):
        result[key] = value
    result.update({
      'options': self.options,
      'project_dir': self.project_dir
    })
    return result


#: Proxy object that points to the current :class:`Session` object.
session = werkzeug.LocalProxy(lambda: Session.current)
